/* messagepassing.c   message passing function calls for the inter-module communication
                      all mp_XXX functions are reentrant (synchronized with semaphores)   
*/

#include <unistd.h>
#include <semaphore.h>
#include "messagepassing.h"
#include "debug.h"
    
/* how the message is stored on the message heap */
typedef struct {
    unsigned char from;
    unsigned char msgType;
    unsigned char length;
    unsigned char *data;
    unsigned char next;
} msg_struct;
    
/* pointers to message handlers for all registered modules */
static message_handler_function message_handlers[MAX_MODULES];

/* immediate[i] is 1, if module i is in immediate mode, 0. if it's not immediate mode
   and 255, if the module doesn't exist */
static unsigned char immediate[MAX_MODULES];

/* message heap */
static msg_struct messages[MAX_MESSAGES];

/* indexes to messages array of the first message in the individual module queues, or 255 if no msg. */
static unsigned char msg_queue[MAX_MODULES];

/* index of the first free block in the message heap */
static unsigned char first_free;

/* # of registered modules */
static unsigned char modcount;

/* message queue semaphores */
static sem_t queue_sem[MAX_MODULES];

/* main semaphore for exclusive access to data structures */
static sem_t mp_sem;

/* --- for descriptions of functions see messagepassing.h --- */

unsigned char mp_register(message_handler_function message_handler, unsigned char immediate_mode)
{
    unsigned char i;
    unsigned char data[2];

    sem_wait(&mp_sem);
    
    for(i = 0; i < MAX_MODULES; i++)
        if (immediate[i] == 255) break;
    
    if (i == MAX_MODULES) {            /* no free module id */
        sem_post(&mp_sem);
        return 255;
    }
    
    immediate[i] = immediate_mode;
    msg_queue[i] = 255;
    modcount++;
    message_handlers[i] = message_handler;
    if (message_handler) {
        data[0] = i;
        data[1] = immediate_mode;
        message_handlers[i](MSG_FROMSYSTEM, MSG_INIT, 2, data);   /* send the init message */
    }
    
    /* initialize message queue semaphore to 0 */
    sem_init(&queue_sem[i], 0, 0);

                                                    //printf("mp_register(%d), queue_sem[%d] = %ld\n",i, i, queue_sem[i].sem_count);

    sem_post(&mp_sem);
    return i;

}

unsigned char mp_goodbye(unsigned char module_id)
{
    int i, j;
    
    sem_wait(&mp_sem);

    if (immediate[module_id] == 255)         /* invalid module_id */
    {
        sem_post(&mp_sem);
        return 0;   
    }
    
    /* discard all messages in the queue */
    i = msg_queue[module_id];
    while (i != 255) 
    {
        j = messages[i].next;
        messages[i].next = first_free;
        first_free = i;
        i = j;
    }

    if (message_handlers[module_id])
        message_handlers[module_id](MSG_FROMSYSTEM, MSG_SHUTDOWN, 0, NULL); /* send shutdown message */
    
    /* wake up a possible blocking receive message return */
    sem_post(&queue_sem[module_id]);
    msleep(20);

    /* dealocate message queue semaphore */
    sem_destroy(&queue_sem[module_id]);
    
    immediate[module_id] = 255;  /* make the module id available */
    
    sem_post(&mp_sem);
    
    return 1;

}

void mp_unlock_mp() 
{
    sem_post(&mp_sem);
}

void mp_lock_mp()
{
    sem_wait(&mp_sem);
}

unsigned char mp_set_mode(unsigned char module_id, message_handler_function message_handler, unsigned char immediate_mode)
{
    int i, j;

    sem_wait(&mp_sem);
    
    if (immediate[module_id] == 255) 
    {
        sem_post(&mp_sem);
        return 0;                   /* invalid module_id */
    }

    message_handlers[module_id] = message_handler;

    if (immediate[module_id] == immediate_mode)        /* no mode change */
    {
        sem_post(&mp_sem);
        return 1;
    }
        
    /* if switching to immediate mode, deliver all messages to the message_handler */
    if (immediate_mode)
    {
        i = msg_queue[module_id];
        while (i != 255)
        {
            message_handler(messages[i].from, messages[i].msgType, messages[i].length, messages[i].data);
            j = messages[i].next;
            messages[i].next = first_free;
            first_free = i;
            i = j;
        }
        msg_queue[module_id] = 255;
    }
    immediate[module_id] = immediate_mode;

    /* send a message to the module that changed mode */
    
    if (message_handler)                                            /* send the init message */
        message_handlers[i](MSG_FROMSYSTEM, MSG_MODECHANGED, 1, (unsigned char *)&immediate[module_id]);

    sem_post(&mp_sem);
    return 1;
}

unsigned char mp_get_mode(unsigned char module_id)
{
    unsigned char i = immediate[module_id];
    
    if (i == 255) return 0;
    else return i;
}

unsigned char mp_send(unsigned char from, unsigned char recipient, unsigned char msg_type, unsigned char length, unsigned char *data)
{
    unsigned char i, j;
    
                                                            //printf("mp_send(%d,%d,%d);\n", from, recipient, msg_type);

    sem_wait(&mp_sem);
    
    if (immediate[recipient] == 255) 
    {
        sem_post(&mp_sem);
        return 0;
    }
    if (immediate[recipient]) 
        message_handlers[recipient](from, msg_type, length, data);            
    else if (first_free == 255)
    {
        sem_post(&mp_sem);
        return 0;
    }    
    else
    {
        /* allocate new block */
        i = first_free;
        first_free = messages[first_free].next;
        /* save message info */
        messages[i].from = from;
        messages[i].msgType = msg_type;
        messages[i].length = length;
        messages[i].data = data;
        messages[i].next = 255;
        /* find the end of the queue and put the message to the queue */
        j = msg_queue[recipient];
        if (j == 255) msg_queue[recipient] = i;
        else
        {
            while (messages[j].next != 255)
                j = messages[j].next;
            messages[j].next = i;
        }
        /* update message queue semaphore (i.e. wakeup the mp_receive if waiting) */
        sem_post(&queue_sem[recipient]);        
	//printf("mp_send(%d,%d,%d); queue_sem[%d] = %ld\n",from,recipient,msg_type, recipient, 0); //queue_sem[recipient].sem_count);
    }
    
    sem_post(&mp_sem);
    return 1;
}

unsigned char mp_receive(unsigned char *from, unsigned char recipient, unsigned char *msg_type, unsigned char *length, unsigned char **data)
{
    unsigned char i;

    if (immediate[recipient] == 255) return 0;

    //printf("mp_receive(%d,%d,%d); queue_sem[%d] = %ld\n",from,recipient,msg_type, recipient, 0 ); //queue_sem[recipient].sem_count);

    /* wait for a message (or KILL signal) */
    sem_wait(&queue_sem[recipient]);

                                                //printf("mp_receive(%d,%d,%d); - message arrived \n",from,recipient,msg_type);

    sem_wait(&mp_sem);
    
    if (immediate[recipient] == 255)     /* check again, somebody could call mp_goodbye meanwhile */
    {
        sem_post(&mp_sem);
        return 0;
    }
    
    i = msg_queue[recipient];
    *from = messages[i].from;
    *msg_type = messages[i].msgType;
    *length = messages[i].length;
    *data = messages[i].data;
    msg_queue[recipient] = messages[i].next;
    /* release block */
    messages[i].next = first_free;
    first_free = i;

                                                //printf("mp_receive(%d,%d,%d); - returning message \n",from,recipient,msg_type);

    sem_post(&mp_sem);

    return 1;
}

unsigned char mp_receive_nb(unsigned char *from, unsigned char recipient, unsigned char *msg_type, unsigned char *length, unsigned char **data)
{
    unsigned char i;
    
    sem_wait(&mp_sem);

    if (sem_trywait(&queue_sem[recipient]) != 0)  /* hopefully doesn't fail with other error than 
                                                    EAGAIN - slightly diff. behav. in legOS and POSIX */
    {
        sem_post(&mp_sem);
        return 0;
    }
    
    i = msg_queue[recipient];
    *from = messages[i].from;
    *msg_type = messages[i].msgType;
    *length = messages[i].length;
    *data = messages[i].data;
    msg_queue[recipient] = messages[i].next;
    /* release block */
    messages[i].next = first_free;
    first_free = i;
    
    sem_post(&mp_sem);
    return 1;
}

unsigned char mp_peek_nb(unsigned char *from, unsigned char recipient, unsigned char *msg_type, unsigned char *length, unsigned char **data)
{
    unsigned char i;
    
    sem_wait(&mp_sem);

    if (sem_trywait(&queue_sem[recipient]) != 0)  /* hopefully doesn't fail with other error than 
                                                    EAGAIN - slightly diff. behav. in legOS and POSIX */
    {
        sem_post(&mp_sem);
        return 0;
    }
    
    i = msg_queue[recipient];
    *from = messages[i].from;
    *msg_type = messages[i].msgType;
    *length = messages[i].length;
    *data = messages[i].data;
    sem_post(&queue_sem[recipient]);
    
    sem_post(&mp_sem);
    return 1;
}

unsigned char mp_clear_buffer(unsigned char module_id)
{
    unsigned char i;
    
    sem_wait(&mp_sem);

    if (immediate[module_id] == 255) 
    {
        sem_post(&mp_sem);
        return 0;                   /* invalid module */
    }
    
    /* release all message blocks */
    i = msg_queue[module_id];
    while (i != 255)
    {
        messages[i].next = first_free;
        first_free = i;
    }
    msg_queue[module_id] = 0;

    /* reinitialize semaphore */
    sem_destroy(&queue_sem[module_id]);
    sem_init(&queue_sem[module_id], 0, 0);
    
    sem_post(&mp_sem);
}

void mp_init()
{
    unsigned char i;
    debugmsg(DL_INFO, "initializing the message passing...\n");

    /* initialize the heap, set all immediate[i] to 255, # of modules to 0. */
    modcount = 0;
    for(i = 0; i < MAX_MODULES; i++)
        immediate[i] = 255;
    for(i = 0; i < MAX_MESSAGES - 1; i++)
        messages[i].next = i+1;
    messages[MAX_MESSAGES - 1].next = 255;
    first_free = 0;
    sem_init(&mp_sem, 0, 1);
}
