/* messagepassing.h   constants and types used by messagepassing.c */

#ifndef __messagepassing_h__
#define __messagepassing_h__

#define MSG_INIT        255         /* sent to the module when registered, len=2, data=module id, immediate_mode(0/1) */
#define MSG_SHUTDOWN    254         /* sent to the module when goodbye is called for the module, len=0 */
#define MSG_MODECHANGED 253            /* sent to the module when the immediate mode is changed by mp_set_mode(), len=1, data=new immediate_mode(0/1) */

#define MSG_BROADCAST   252         /* message send to all registered modules */
#define MSG_DEFAULT     0           /* message send to a default module (the first registered) */
#define MSG_FROMSYSTEM  251         /* from field of the message in system messages */

#define MAX_MODULES     20          /* maximum number of modules */

#define MAX_MESSAGES    200         /* maximum number of messages in the queue */

#define IMMEDIATE_MODE  1           /* mp_register constant for argument immediate_mode */
#define QUEUE_MODE      0           /* mp_register constant for argument immediate_mode */

/* pointer to message_handler function. Message handler gets the id of the sending module,
   message type, number of bytes and the pointer to raw byte data that are stored 
   either where the sending module had the message (it's not copied) if in the same robot,
   or in the message passing buffers, in that case it won't exist after this function returns. */
typedef void (*message_handler_function) (unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data);

/* function prototypes */

/*  registers new module that will be sending/receiving messages
     set immediate_mode to 0, if the module will receive messages with peek/receive/receive_all
     set immediate_mode to 1, if the module's message_handler should be called
      immediatelly when the message arrives. This handler will be called in the
      thread of the calling process, and the corresponding mp_send() will return only after
      the message handler returns, so it should return quickly. The handler normally cannot 
      send messages to other modules (this will cause deadlock). If it's really needed, it 
      should first call mp_unlock_mp(), then send the message (but risking that the handler 
      will be called again - if that can happen, programmer of the module should make sure 
      that the message handler is reentrant), and after sending the message, the handler should
      call mp_lock_mp() again.
     if immediate_mode is 0, message_handler can be NULL.
     if message_handler is not NULL (regardless immediate_mode), it's sent a
      system meesage with msgtype set to MSG_INIT and the module ID - single unsigned char 
      followed by immediate_mode (unsigned char) in the data (2 bytes).
     returns id (>=0) of the module (unique for this robot) or 255 if # of modules exceeded. */
unsigned char mp_register(message_handler_function message_handler, unsigned char immediate_mode);

/* removes the module from the list of modules that receive/send messages. 
   All messages for this module waiting in the queue are discarded. If the message handler of
   the module is not NULL, it will be sent the SHUTDOWN message through the message handler
   and if there is a mp_receive (or similar) waiting for a message for this module, it will
   return with 0 (I believe).
   returns 0, if module_id is not valid, otherwise returns 1. */
unsigned char mp_goodbye(unsigned char module_id);

/* unlocks the message passing semaphore. Should be called from immediate mode message handler,
   which wants to call any mp_XXX routine immediatelly before that call. mp_lock_mp() must be 
   called immediatelly after such call */
void mp_unlock_mp();

/* locks the message passing semaphore back. Should be called only after mp_unlock_mp() */
void mp_lock_mp();

/* allows to change the immediate mode for a given module. Messages waiting in the queue for
   a non-immediate mode module will be delivered to the message_handler immediatelly.
   if the new immediate_mode is the same as current, at least message_handler is updated.
   if the mode is changed, MSG_MODECHANGED is sent to the new messagehandler
   returns 0, if module_id is not valid, otherwise returns 1.
   --if you are not using this function in the robot, you can just comment it out-- */
unsigned char mp_set_mode(unsigned char module_id, message_handler_function message_handler, unsigned char immediate_mode);

/* returns 1, if the module is in immediate mode, otherwise returns 0. */
unsigned char mp_get_mode(unsigned char module_id);

/* sends a message to a recipient module. If the recipient is in immediate mode,
   the function will return after recipient will process the message and it's message 
   handler returns. If the recipient is not in immediate mode, the message will be added
   to the queue of messages for the recipient and delivered when it calls the receive/recieve_all
   function. Since only a pointer to the data is passed, when the recipient is not in
   immediate mode, it will probably access the data after the mp_send returns! The calling
   module should therefore either send a pointer to some constant message table or keep
   the data long enough unmodified in order to let the recipient use them. :-(( 
   Therefore you might want to have your module running in immediate mode even if it's
   running its own thread and copy the message data by its own message handler...
   This solution was chosen in order to avoid copying of data when it's not necessary, but
   the module programmers should pay attention to this potential source of nasty bugs.
   Please also note that if you are using a semaphore inside of a message handler, the
   corresponding critical regions in other threads, which are locking this semaphore
   may not call mp_send, otherwise deadlock can occur.
   returns 0, if queue is full or recipient unknown, 1 otherwise. */
unsigned char mp_send(unsigned char from, unsigned char recipient, unsigned char msg_type, unsigned char length, unsigned char *data);

/* receives a message if there is one waiting in the queue. Should be used by modules
   running in non-immediate mode. If no message is waiting, this call blocks, until 
   a message arrives. The message is saved to variables refered to by arguments, 
   *data will contain a pointer to the first byte of the continuous message data. 
   Can be callled even if the recipient is in immediate mode, but will block until 
   some other thread will switch the recipient to non-immediate mode and send it a message
   returns 0, if the recipient is invalid, otherwise returns 1 */
unsigned char mp_receive(unsigned char *from, unsigned char recipient, unsigned char *msg_type, unsigned char *length, unsigned char **data);

/* receives a message if there is one waiting in the queue. If no message, returns 0, otherwise returns 1.
   the message is saved to parameters, data will contain a pointer to a message
   the message is deleted from the queue. */
unsigned char mp_receive_nb(unsigned char *from, unsigned char recipient, unsigned char *msg_type, unsigned char *length, unsigned char **data);

/* receives a message if there is one waiting in the queue. If not, returns 0, otherwise returns 1.
   the message is saved to parameters, data will contain a pointer to a message
   the message is not deleted from the queue. */
unsigned char mp_peek_nb(unsigned char *from, unsigned char recipient, unsigned char *msg_type, unsigned char *length, unsigned char **data);

/* clears the message queue of a given module 
   returns 0, if module_id is not valid, otherwise returns 1
   note: I'm not sure if this can be called while some other thread is waiting for a message,
         because the semaphore that it is waiting on will be reinitialized... */
unsigned char mp_clear_buffer(unsigned char module_id);

/* initializes message passing data structures, should be called before a first call to mp_XXX function; 
   should be called by one thread only */
void mp_init();


#endif    // __messagepassing_h__