/* modules.c   Set of basic modules for robot control, see modules.h
                you might want to uncomment those parts of message handlers which you
                don't use in your project to save LEGO memory...
*/

#include "modules.h"
#include "debug.h"

#ifdef RUN_IN_SIMULATION

#include <stdio.h>
#include <semaphore.h>
#include "simulator.h"

#endif

#ifdef RUN_ON_REAL_ROBOT

#include <dsensor.h>
#include <dmotor.h>
//...?

#endif

#include "messagepassing.h"

/* module Navigate */

#define NAVIGATE_NO_COMMAND        127        // internal command id, no new command arrived

static pid_t navigate_pid;              // navigate's own thread's pid
static unsigned char next_cmd;          // message that arrived and wasn't fetched yet
static unsigned char next_arg;            // and its argument
static unsigned char *next_finished_flag;    // pointer to a variable where the navigate 
                                        // module will indicate the status of the arrived
                                        // command (message)
static unsigned char navigate;          // navigate thread running
static time_t timeout;                    // when the command will be finished 
static unsigned char movedata[3];        // data for message sent to motordriver
static unsigned char offdata[3];        // the same for breaking motors 
static unsigned char nopowerdata[2];    // the same for turning breaks off

static sem_t nav_sem;                    // semaphore for coordination of navigation thread and message handler

static unsigned char garbage;            // where the flag is saved when not needed

wakeup_t navigate_execute_wakeup(wakeup_t i)
{
    return ((sys_time >= timeout) || (next_cmd != NAVIGATE_NO_COMMAND));
}

wakeup_t navigate_newmsg_wakeup(wakeup_t i)
{
    return (next_cmd != NAVIGATE_NO_COMMAND);
}

int navigate_main(int argc, char **argv)
{
    unsigned char motorsoff;        // should the motors be turned off?
    unsigned char cmd;              // message currently processed
    unsigned char arg;                // and its argument
    unsigned char *finished_flag;     // pointer to a variable where the navigate 
                                    //  module indicates the status of the current cmd

	//printf("module_navigate_main\n");

    while (navigate) {       // while not SHUTDOWN

        // wait for the command to arrive
        
                    // printf("waiting for a message\n");
        wait_event(navigate_newmsg_wakeup, 0);

        // fetch new message
        
        // prevent message handler from modifying data while fetching newly arrived message...
                    //printf("fetching the message\n");
        sem_wait(&nav_sem);    

            cmd = next_cmd;
            arg = next_arg;
            finished_flag = next_finished_flag;
            next_cmd = NAVIGATE_NO_COMMAND;

        sem_post(&nav_sem);

        //execute fetched message            
        
        // make sure the power is positive (last command could have left it negative)        
                    //printf("executing the message\n");
        movedata[1] = abs(movedata[1]);        
        if (cmd == NAVIGATE_RIGHT) 
            // if turning, set the direction
            CHANGE_SIGN_UNSIGNED_CHAR(movedata[1]);            
        
        // compute command duration
        timeout = sys_time + arg * NAVIGATE_TIMESTEP;
        // by default command needs to turn motors off when finished
        motorsoff = 1;        

        switch (cmd)
        {
            case NAVIGATE_FWD:
                movedata[0] = 5;
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_POWER, 3, movedata);
                break;
            case NAVIGATE_LEFT:
            case NAVIGATE_RIGHT:
                movedata[0] = 1;
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_POWER, 3, movedata);
                movedata[0] = 4;                
                CHANGE_SIGN_UNSIGNED_CHAR(movedata[1]);
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_POWER, 3, movedata);
                break;
            case NAVIGATE_TURNRND:
                //here, timeout should be also random
                timeout = sys_time + (random() % NAVIGATE_MAX_RND_TIME) * NAVIGATE_TIMESTEP;
                if (random() & 1) movedata[0] = 4;
                else movedata[0] = 1;
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_POWER, 3, movedata);
                movedata[0] = 5 - movedata[0];
                CHANGE_SIGN_UNSIGNED_CHAR(movedata[1]);
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_POWER, 3, movedata);
                break;
            case NAVIGATE_STOP:                
                timeout = sys_time; // command finished now, motors will be turned off bellow
                break;
            case NAVIGATE_FAST:                
            case NAVIGATE_SLOW:
                movedata[1] = (cmd == NAVIGATE_SLOW) ? 63 : 127;
                movedata[0] = 5;
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_SETPOWER, 3, movedata);
                timeout = sys_time; // command finished now
                motorsoff = 0;        // don't turn motors off please
                break;
        }

        // wait until command finishes or a new command arrives
                    //printf("waiting until finished or new message timeout=%lu\n sys_time=%lu", timeout, sys_time);
        wait_event(navigate_execute_wakeup, 0);

        //if command didn't arrived, -> command finished
        if (next_cmd == NAVIGATE_NO_COMMAND)
        {
                        //printf("turning off motors if necessary\n");
            // brake motors if necessary
            if (motorsoff)
            {
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_POWER, 3, offdata);
                timeout = sys_time + MOTORDRV_BRAKE_TIME;           
    
                // wait while motors stop or message arrives
                            //printf("waiting while motors stop or message comes\n");
                wait_event(navigate_execute_wakeup, 0);

                // make the motors free 
                            //printf("making the motors free\n");
                mp_send(mid_navigate, mid_motordrv, MOTORDRV_NOPOWER, 3, nopowerdata);
            }

                            //printf("setting the finished flag\n");
            *finished_flag = NAVIGATE_COMMAND_FINISHED;
        }
        else
        {
                            //printf("setting the finished flag to interrupted\n");
            *finished_flag = NAVIGATE_COMMAND_INTERRUPTED;
        }
        
    } //while(timeout)
}

void navigate_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data)
{
    switch (msgType)
    {
        case MSG_INIT:
            //printf("navigate_message MSG_INIT\n");
            // start navigate module thread
            movedata[0] = 0; movedata[1] = 127; movedata[2] = NAVIGATE_MOTOR_PRIORITY;
            offdata[0] = 5; offdata[1] = 0; offdata[2] = NAVIGATE_MOTOR_PRIORITY;
            nopowerdata[0] = 5; nopowerdata[1] = NAVIGATE_MOTOR_PRIORITY;    
            next_cmd = NAVIGATE_NO_COMMAND;
            next_finished_flag = &garbage;
            sem_init(&nav_sem, 0, 1);
            navigate = 1;
            //printf("calling execi(navigate_main)\n");
            //printf("navigate_main : %ld\n", navigate_main);
            navigate_pid = execi(navigate_main, 0, NULL, PRIO_NORMAL, DEFAULT_STACK_SIZE);            
            //printf("navigate_pid:%d\n",navigate_pid);
            break;

        case MSG_SHUTDOWN:
            // terminate navigate module thread
            sem_wait(&nav_sem);
              next_finished_flag = &garbage;
              next_cmd = NAVIGATE_STOP;
              navigate = 0;
            sem_post(&nav_sem);
            break;

        // all the navigate messages have the same format
        case NAVIGATE_FWD: 
        case NAVIGATE_LEFT:
        case NAVIGATE_RIGHT:
        case NAVIGATE_TURNRND: 
        case NAVIGATE_STOP: 
        case NAVIGATE_FAST: 
        case NAVIGATE_SLOW: 

            // prevent navigate thread from fetching an incomplete message
            sem_wait(&nav_sem);    
                                                        
                // remember the new command and its argument 
                // until the execution thread doesn't pick it up
                next_arg = data[NAVIGATE_N];
                next_cmd = msgType;
                // copy the pointer stored in data to unsigned char var.              
                next_finished_flag = *((unsigned char **)(&data[NAVIGATE_FINISHED]));    
                if (next_finished_flag == NULL) next_finished_flag = &garbage;
                *next_finished_flag = NAVIGATE_COMMAND_STILLACTIVE;                
            
            sem_post(&nav_sem);    // ok, navigate thread go ahead!
            
            break;
    }
}

/* Module Cargo */

static pid_t cargo_pid;              // cargo's own thread's pid

int cargo_main(int argc, char **argv)
{
    unsigned char cargo_load[3] = { 2, 127, CARGO_MOTOR_PRIORITY };
    unsigned char msgType, length, from;
    unsigned char *data;

    while(mp_receive(&from, mid_cargo, &msgType, &length, &data)) 
    {    
        switch (msgType)    
        {
          case MSG_SHUTDOWN: 
              return 0;
          case CARGO_LOAD: 
            cargo_load[1] = 127;
            break;
          case CARGO_UNLOAD:
            cargo_load[1] = -127;
            break;
          default:
            continue;
        }

        cargo_load[1] = 127;
        mp_send(mid_cargo, mid_motordrv, MOTORDRV_POWER, 3, cargo_load);
        msleep(500);
        cargo_load[1] = 0;
        mp_send(mid_cargo, mid_motordrv, MOTORDRV_POWER, 3, cargo_load);
        msleep(50);
        mp_send(mid_cargo, mid_motordrv, MOTORDRV_NOPOWER, 3, cargo_load);        
    }    
}

void cargo_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data)
{
    switch (msgType)
    {
        case MSG_INIT:
            // start cargo module thread
            cargo_pid = execi(cargo_main, 0, NULL, PRIO_NORMAL, DEFAULT_STACK_SIZE);
            break;

/*        case MSG_SHUTDOWN:
            // terminate cargo module thread
            mp_send(from, mid_cargo, MSG_SHUTDOWN, 0, NULL);
            break; */
    }            
}

/* Module Line follower */
   
static pid_t lnflwer_pid;              // lnfollower's own thread's pid

int lnflwer_main(int argc, char **argv)
{
    unsigned char msgType, length, from;
    unsigned char *data;
    unsigned char turn[3] = { 1, 127, LNFLWER_MOTOR_PRIORITY };
    unsigned char fwd[3] = {5, 127, LNFLWER_MOTOR_PRIORITY };
    int timeout, total_timeout;

    // wait for a message FOLLOW
    while (mp_receive(&from, mid_lnflwer, &msgType, &length, &data))
    {      
      total_timeout = sys_time + LNFLWER_TIMEUNIT * (*data);

      if (msgType != LNFLWER_FOLLOW) continue;      
        
      while (sys_time < total_timeout) {
        // move forward        
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, fwd);

        // line?
        while (LIGHT_2() < LNFLWER_LINETHRESHOLD) 
        {
          // check for a new message
          if (0 != mp_peek_nb(&from, mid_lnflwer, &msgType, &length, &data))
          { 
              total_timeout = sys_time - 1;  // arrived -> we should finish now
              break;
          }
          // also check timeout of FOLLOW command
          if (sys_time >= total_timeout) break;
          // if still line -> keep going
        }        
        
        // new message or timeout? -> stop motors and get new message
        if (sys_time >= total_timeout) break;

        // line discontinued ->
        
        // first check for a new message also here (may be there is no line)
        if (0 != mp_peek_nb(&from, mid_lnflwer, &msgType, &length, &data))
          break;
                
        // and start turning
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, turn);
        turn[0] = 5 - turn[0];
        turn[1] = - turn[1];
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, turn);  
               
        timeout = sys_time + LNFLWER_TURNTIMEOUT;
 
        // line?
        while (LIGHT_2() >= LNFLWER_LINETHRESHOLD)
        { // no -> timelimit?
          if (sys_time > timeout) 
          {
            timeout = 1;
            break; // yes -> see below...
          }
          // no -> keep turning
        }

        // found line?
        if (timeout != 1)
          continue;   //yes -> now follow it

        // no (was timelimit) -> turn opposite
        turn[1] = - turn[1];
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, turn); 
        turn[0] = 5 - turn[0];
        turn[1] = - turn[1];
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, turn); 

        timeout = sys_time + 2 * LNFLWER_TURNTIMEOUT;

        // line?
        while (LIGHT_2() >= LNFLWER_LINETHRESHOLD)
        { // no -> timelimit?
          if (sys_time > timeout)
          {
            timeout = 1;
            break;
          }
          // no -> keep turning
        }

        // found line?
        if (timeout != 1)
          continue;  // yes -> now follow it

        // no (was timelimit) -> turn back              
        turn[1] = - turn[1];
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, turn);
        turn[0] = 5 - turn[0];
        turn[1] = - turn[1];
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, turn);

        msleep(LNFLWER_TURNTIMEOUT);

        // move forward a bit
        mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, fwd);
        msleep(LNFLWER_FWDABITTIMEOUT);
        // and try again
      }

      // brake and stop motors
      mp_send(mid_lnflwer, mid_motordrv, MOTORDRV_POWER, 3, offdata);
      msleep(MOTORDRV_BRAKE_TIME);
      mp_send(mid_navigate, mid_motordrv, MOTORDRV_NOPOWER, 3, nopowerdata);
    }
}

void lnflwer_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data)
{
    switch (msgType)
    {
        case MSG_INIT:
            // start lnflwer module thread
            lnflwer_pid = execi(lnflwer_main, 0, NULL, PRIO_NORMAL, DEFAULT_STACK_SIZE);
            break;
    }
}

/* Module Motordriver */

/* garbage (e2) 
typedef void (*motor_dir_function) (MotorDirection);
typedef void (*motor_speed_function) (unsigned char);

motor_dir_function motor_X_dir[3] = { motor_a_dir, motor_b_dir, motor_c_dir };
motor_speed_function motor_X_speed[3] = { motor_a_speed, motor_b_speed, motor_c_speed };
*/

void motordriver_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data)
{
    static unsigned char priority[3] = { MOTORDRV_MIN_PRIORITY, 
                                         MOTORDRV_MIN_PRIORITY, 
                                         MOTORDRV_MIN_PRIORITY };
    unsigned char motor, prior;
    char power;
    MotorDirection motordir;

    switch (msgType)
    {
        case MOTORDRV_SETPOWER:
            motor = data[0];
            power = (char)(data[1]) << 1;
            prior = data[2];

            if ((motor & 1) && (prior >= priority[0]))
            {
                    motor_a_speed(power);
                    priority[0] = prior;
            }
            if ((motor & 2) && (prior >= priority[1]))
            {
                    motor_b_speed(power);
                    priority[1] = prior;
            }
            if ((motor & 4) && (prior >= priority[2]))
            {
                    motor_c_speed(power);
                    priority[2] = prior;
            }
            break;

        case MOTORDRV_NOPOWER:
            motor = data[0];
            prior = data[1];
            if ((motor & 1) && (prior >= priority[0]))
            {
                motor_a_dir(off);
                priority[0] = prior;
            }
            if ((motor & 2) && (prior >= priority[1]))
            {
                motor_b_dir(off);
                priority[1] = prior;
            }
            if ((motor & 4) && (prior >= priority[2]))
            {
                motor_c_dir(off);
                priority[2] = prior;
            }
            break;

        case MOTORDRV_POWER:
            motor = data[0];
            power = (char)(data[1]);
            prior = data[2];

            if (power > 0) motordir = fwd;
            else if (power < 0)
            {
                motordir = rev;
                power = -power;
            }
            else motordir = brake;
            power <<= 1;

            if ((motor & 1) && (prior >= priority[0]))
            {
                    motor_a_dir(motordir);
                    motor_a_speed(power);
                    priority[0] = prior;            
            }
            if ((motor & 2) && (prior >= priority[1]))
            {
                    motor_b_dir(motordir);
                    motor_b_speed(power);
                    priority[1] = prior;
            }
            if ((motor & 4) && (prior >= priority[2]))
            {
                    motor_c_dir(motordir);
                    motor_c_speed(power);
                    priority[2] = prior;
            }
            break;
        case MOTORDRV_SAVESTATUS:
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! TODO
            break;
        case MOTORDRV_RESTORESTATUS:
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! TODO
            break;
    }
}

/* Module Avoidance */

static pid_t avoidance_pid;              // lnfollower's own thread's pid

int avoidance_main(int argc, char **argv)
{
    unsigned char msgType, length, from;
    unsigned char *data;
    unsigned char stopdata[3] = { 5, 0, AVOIDANCE_MOTOR_PRIORITY };
    unsigned char backup[3] = {5, -127, AVOIDANCE_MOTOR_PRIORITY };
    unsigned char turn[3] = { 1, 127, AVOIDANCE_MOTOR_PRIORITY };    
    unsigned char saved_motor_status[MOTORDRV_MOTORSTATUS_SIZE];
    unsigned char stop_message;

    // wait for a message START
    while (mp_receive(&from, mid_avoidance, &msgType, &length, &data))
    {      
      if (msgType != AVOIDANCE_START) continue;
      stop_message = 0;
      while (!stop_message)
      {
        // wall?
        while (SENSOR_1() > AVOIDANCE_WALLTHRESHOLD) 
        {
          // check for a new message
          if (0 != mp_peek_nb(&from, mid_avoidance, &msgType, &length, &data))
          { 
             stop_message = 1;              
             break;
          }          
        // if still free -> keep waiting
        }
        
        // new message? -> get a new one
        if (stop_message) break;
        
        // save current motor status        
        mp_send(mid_avoidance, mid_motordrv, MOTORDRV_SAVESTATUS, 0, saved_motor_status);
        
        // brake motors
        mp_send(mid_avoidance, mid_motordrv, MOTORDRV_POWER, 3, stopdata);
        msleep(MOTORDRV_BRAKE_TIME);
        
        // obstacle -> backup
        mp_send(mid_avoidance, mid_motordrv, MOTORDRV_POWER, 3, backup);
        msleep(AVOIDANCE_BACKUP_TIME);

        // now turn random direction
        if (sys_time & 1) turn[1] = - turn[1];
        mp_send(mid_avoidance, mid_motordrv, MOTORDRV_POWER, 3, turn);
        turn[0] = 5 - turn[0];
        turn[1] = - turn[1];
        mp_send(mid_avoidance, mid_motordrv, MOTORDRV_POWER, 3, turn);
        msleep(AVOIDANCE_TURN_TIME);

        // restore previous motor status
        mp_send(mid_avoidance, mid_motordrv, MOTORDRV_RESTORESTATUS, 0, saved_motor_status);
        
        // and wait again
      }
    }
}


void avoidance_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data)
{
    switch (msgType)
    {
        case MSG_INIT:
            // start lnflwer module thread
            avoidance_pid = execi(avoidance_main, 0, NULL, PRIO_NORMAL, DEFAULT_STACK_SIZE);
            break;
    }
}

/* Module Console */

void console_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data)
{
    switch (msgType)
    {
        case MSG_INIT: printf("\nCONSOLE MODULE INITIALIZED. WILL PRINT IN THE FOLLOWING FORMAT:\n[from_module;time]:message_text\t\t(note: 251 is from system)\n\n");
             break;
        case CONSOLE_PRINT: printf("[%d;%d]:%s\n", from, (int)(sys_time % (unsigned long)600000), data);
                            break;
    }
}

// print string on a console
void writeln(char *ln)
{
    mp_send(MSG_FROMSYSTEM, mid_console, CONSOLE_PRINT, strlen(ln), ln);
}
