/* modules.h   message constants and function prototypes for basic set of modules 
               note: message type constants should be unique only within a module,
                     even though they are made unique over all the moduls */

#ifndef __modules_h__
#define __modules_h__

/* This source code should be linked either together with simulator.c or with LegOS, uncomment
the appropriate: */

#define RUN_IN_SIMULATION
//#define RUN_ON_REAL_ROBOT


/* ----- message constants */

/* module Navigate    (should be registered in immediate, with navigate_message handler, runs its own thread) 

      fwd n     start moving n small steps forward, next received message aborts
      right n   start turning right n small steps, next message aborts
      left n    start turning left n small steps, next message aborts
      turnrnd n start turning in random direction random number of steps (up to 180 deg.)
                  idea is to use this for env. exploration, n is ignored
      stop n    stop moving, n is ignored
      fast n    switch robot to a fast speed mode (immediate effect), n is ignored
      slow n    switch robot to a slow speed mode (immediate effect), n is ignored

      each command takes an extra argument *var, which is a pointer to an unsigned char
      variable. Set var to NULL, if you don't care, otherwise this variable is set
      to NAVIGATE_COMMAND_FINISHED when the command finishes normally and to 
      NAVIGATE_COMMAND_INTERRUPTED when it finishes prematurely because of another 
      navigate command received. This variable is initialized to
      NAVIGATE_COMMAND_STILLRUNNING before returning from mp_send()
*/

// on RISC processors, pointer can be stored only to a (4k+0)-type memory address
#ifdef RUN_IN_SIMULATION
#define platform_align(x)    ((x + 15) & (~15))
#else
#define platform_align(x)    x
#endif

// navigate message types (commands)
#define NAVIGATE_FWD        11
#define NAVIGATE_RIGHT        12
#define NAVIGATE_LEFT        13
#define NAVIGATE_TURNRND    14
#define NAVIGATE_STOP        15
#define NAVIGATE_FAST        16
#define NAVIGATE_SLOW        17

#define NAVIGATE_SHUTDOWN   18

#define NAVIGATE_MOTOR_PRIORITY     4

// navigate message argument indexes
#define NAVIGATE_N            0
#define NAVIGATE_FINISHED    platform_align(1)

// length of navigate messages in bytes
#define NAVIGATE_MSGSIZE    NAVIGATE_FINISHED + sizeof(void *)

// constants for finished flag
#define NAVIGATE_COMMAND_STILLACTIVE        0
#define NAVIGATE_COMMAND_FINISHED            1
#define NAVIGATE_COMMAND_INTERRUPTED        2

// other and internal defines for navigate module
#define NAVIGATE_TIMESTEP    100            // duration (ms) of one navigate time unit
#define NAVIGATE_MAX_RND_TIME    15        // maximum number of time units for random turns
#define MAXTIME                0x7FFFFFFF    // later than any other system time
#define CHANGE_SIGN_UNSIGNED_CHAR(x)    x = (unsigned char)(0 - (char)x)                

/* Module Cargo        (should be registered in non-immediate, with cargo_message handler)

      load      to load cargo when at cargo loading station
      unload       to unload cargo when at cargo unloading station

*/

#define CARGO_MOTOR_PRIORITY    5

#define CARGO_LOAD            21
#define CARGO_UNLOAD        22

/* Module Line follower        (should be registered in non-immediate mode, with lnflwer_message handler)

      follow n  will start following line at most n small steps, if 
                   the line is detected
      stop      stops following the line

*/

#define LNFLWER_MOTOR_PRIORITY  3

#define LNFLWER_FOLLOW        31
#define LNFLWER_STOP        32

#define LNFLWER_LINETHRESHOLD   35          // readings on light sensor below this value are interpreted as line
#define LNFLWER_TIMEUNIT        100         // duration of 1 small step
#define LNFLWER_TURNTIMEOUT     500         // for how long time does the robot turn to one direction to search for a line
#define LNFLWER_FWDABITTIMEOUT  300         // how much to travel forward when line not detected on any side

/* Module Motordriver        (should be registered in immediate mode with motordriver_message handler)

        power MOTOR SPEED PRIORITY        sets the power of the motor 
                                        (A,B,C) - encoded as 1,2,4 to 
                                        POWER \in \{-127,127\}, 
                                        if the priority is higher or equal to 
                                        current priority, 0 means break.
        nopower MOTOR PRIORITY            puts motor into loose mode (i.e. no active break)

        setpower MOTOR SPEED PRIORITY    updates the speed of specified motors to a given value without
                                        changing their direction

        savestatus WHERE_TO_SAVE        overwrites data with current status of motors 
        restorestatus SAVED_STATUS      restores previously saved status of motors

   assumes 1 centered light sensor on port 2 
   in front of the robot just above the surface pointing downwards 

*/

#define MOTORDRV_POWER            41
#define MOTORDRV_NOPOWER        42
#define MOTORDRV_SETPOWER        43
#define MOTORDRV_SAVESTATUS     44
#define MOTORDRV_RESTORESTATUS  45
#define MOTORDRV_MOTORSTATUS_SIZE 3 * sizeof(MotorState)

#define MOTORDRV_MIN_PRIORITY    0
#define MOTORDRV_MAX_PRIORITY    255

#define MOTORDRV_BRAKE_TIME     70            // for how long time (ms) are the breaks on 


/* Module Avoidance        (should be registered in non-immediate, with avoidance_message handler)

      start        start the obstacle avoidance behavior
      stop      stop the obsatacle avoidance behavior

    assumes touch sensor bumper on port 1

*/

#define AVOIDANCE_START            51
#define AVOIDANCE_STOP          52

#define AVOIDANCE_WALLTHRESHOLD  50         // sensitivity of the bumper
#define AVOIDANCE_BACKUP_TIME   300         // after running into obstacle, robot backs up 
#define AVOIDANCE_TURN_TIME     300         // after backing up, robot turns in random direction
    
#define AVOIDANCE_MOTOR_PRIORITY 10

/* Module Console    (debugging module for printing debugging information on the console,
                     should be registered in immediate mode with console_message handler)

      print STRING    prints the string on the console
*/

#define CONSOLE_PRINT            61

/* ----- function prototypes */

void navigate_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data);
void cargo_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data);
void lnflwer_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data);
void motordriver_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data);
void avoidance_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data);
void console_message(unsigned char from, unsigned char msgType, unsigned char length, unsigned char *data);
void writeln(char *ln);

/* module identifiers - call to mp_register should store the identifiers in these
   variables, because they are used by the modules !!! */

unsigned char mid_navigate, mid_cargo, mid_motordrv, mid_lnflwer, mid_avoidance, mid_console;    

#endif  // __modules_h__