/* simulator.c    simulates the real robot and environment for the cotroller, provides fitness
                  for the EA 
                  
                  The simulator updates a speed vector based on the changes of the motor
                  power. Each time something happens (either sensors are read, or motor
                  values change, or some other event occurs, the simulation is continued
                  to the current time (which is stopped for the moment of computation)
                  which is scaled so that the simulator will get "realistic" performance.
                     
   Implementation notes:
      Data structures are static arrays. 
      Consider dynamically allocated arrays if lack of memory.
*/

/* CONTROLLER - SIMULATOR interface */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/timeb.h>
#include <math.h>

#include "simulator.h"
#include "simulator_internals.h"
#include "debug.h"

/* global static variables */

/* robot speed vector */
       int robot_moving;		/* one of MOVING_XXX constants */
	   double robot_heading;	/* in radians, 0 is towards higher y coordinate, increasing in clockwise dir. */
	   double robot_speed;		/* simulation distance units per simulation time unit, nonnegative */
static double robot_c_x;		/* center of robot's circular trajectory, x coordinate */
static double robot_c_y;		/* 										  y coordinate */
static double robot_c_r;		/* radius of robot's circular trajectory (to the center of robot) */

/* type of environment */
static int envtype;

/* size of rectangular environment */
static double env_w, env_h;

/* robot location and initial location and heading */
	   double robot_x, robot_y;				/* 0,0 is bottom left corner */
static double init_x, init_y, init_heading;

/* robot dimensions */
	   double robot_radius;

/* robot speed ratio - how is the motor speed translated to robot speed */
static double robot_speed_ratio;

/* simulation time variables */
static long simulation_time;			/* simulation time at last event */
	   long new_simulation_time;        /* simulation time at this event */
static clock_t system_time;		        /* system time at last event */
static clock_t new_system_time;         /* system time at this event */
static double slowdown_constant;        /* for scaling the system time towards simulation time */

/* obstacles */
	   int obstcnt;
	   double obs_x[MAX_OBSTCNT];
	   double obs_y[MAX_OBSTCNT];
	   double obs_w[MAX_OBSTCNT];
	   double obs_h[MAX_OBSTCNT];

static double next_obstacle_x;			/* next point where the robot will run into */						
static double next_obstacle_y;			/* an obstacle or a wall given its current speed vector */

/* floor marks */
	   int markcnt;
	   double mark_x[MAX_MARKCNT];
	   double mark_y[MAX_MARKCNT];
	   double mark_w[MAX_MARKCNT];
	   double mark_h[MAX_MARKCNT];
	   int mark_value[MAX_MARKCNT];

/* light sources */
       int lightcnt;
       double light_x[MAX_LIGHTCNT];
       double light_y[MAX_LIGHTCNT];
       double light_intensity[MAX_LIGHTCNT];
       int light_state[MAX_LIGHTCNT];		/* 1: on, 0: off */

/* active areas */
       int activecnt;
       double active_x[MAX_ACTIVECNT];
       double active_y[MAX_ACTIVECNT];
       double active_w[MAX_ACTIVECNT];
       double active_h[MAX_ACTIVECNT];
static int active_activation[MAX_ACTIVECNT];		// one of the ACTIVE_ONE_TIME_ONLY, ACTIVE_ON_EACH_ENTRANCE, ACTIVE_WITH_DELAY
static int active_delay[MAX_ACTIVECNT];				// used only for ACTIVE_WITH_DELAY
static unsigned int active_last_fired_at[MAX_ACTIVECNT];		// if ACTIVE_WITH_DELAY, simulation time when last activated, otherwise 0 if not activated and 1 if activated
static int active_script_event[MAX_ACTIVECNT];		// index of script to be executed

/* global time events */
static int eventcnt;
static int event_periodic[MAX_EVENTCNT];
static unsigned long event_start[MAX_EVENTCNT];
static int event_count[MAX_EVENTCNT];
static unsigned long event_time[MAX_EVENTCNT];
static int event_script_event[MAX_EVENTCNT];

/* local time events to occur next - these handle the active areas, global time events, and next collision */
static time_event_str time_events[MAX_LOCALTIMEEVENTS];		
static int te_len;

/* event scripts for active areas and time events */
static int scriptcnt;
static int script[MAX_EVENT_SCRIPTS][MAX_SCRIPT_SIZE][2];

/* reads line from file if comment, reads again, if EOF, sets readline_eof flag */
static int readline_eof;

/* statistics variables */
static int num_hits;				/* how many times robot ran into the wall or an obstacle */
static double total_distance;		/* the total distance travelled by the robot */

/* other internal variables */
static int was_by_obstacle;		/* is 1 if at the last simulator call the robot was touching an obstacle */
static double last_robot_heading;	/* this helps to determine whether we have to update the local time */
static double last_robot_speed;		/* events when a simulator function is called. if robot status is the */
static int last_robot_moving, last_robot_c_r, last_robot_c_x, last_robot_c_y;	/* same as last time, we can keep the previous time events */

static int viewer_on;			/* show simulation viewer? */

static int can_keep_events;   /* even the first time event was not reached */

/* description of moving states - just for debugging purposes */
       char *moving_desc[10] = { "", "stopped", "fwd_straight", "bwd_straight",
								 "fwd_left", "bwd_left", "fwd_right", "bwd_right",
								 "rot_left", "rot_right" };

/* here the trajectory of the robot is stored. it is used by viewer and finally saved into a file */
		int	traj_moving[MAX_TRAJECTORY_LEN];
		double traj_x[MAX_TRAJECTORY_LEN];
		double traj_y[MAX_TRAJECTORY_LEN];
		double traj_c_r[MAX_TRAJECTORY_LEN];
		double traj_c_x[MAX_TRAJECTORY_LEN];
		double traj_c_y[MAX_TRAJECTORY_LEN];
		int	traji;

static char trajectory_file[100];	/* path to the file where trajectory is saved */

/*--------------------------------------------------------------------------------------
 sensor handling (equivalent to dsensor.h, has to be updated to new version of LegOS) 
 
 these are all the functions which the controller uses to access the sensors and
 they are equivalent to a real LegOS

  assuming 
    sensor 1 is a front bumper,
    sensor 2 is a light sensor pointing upwards that returns ambient light based
             on a light model
    sensor 3 is a light sensor pointing downwards reading the marks on the floor

  and sensor 1 to be a front bumper
 
*/

int SENSOR_1_val, SENSOR_2_val, SENSOR_3_val, LIGHT_1_val, LIGHT_2_val, LIGHT_3_val,
    ROTATION_1_val, ROTATION_2_val, ROTATION_3_val;

int SENSOR_1() 
{ 
	int rv;
	debugmsg(DL_DEBUGGING, "---B--S1()-\n");
	update_simulation(); 
	debugmsg1long(DL_SIMULATOR, "[sim;%lu]:SENSOR_1()\n", new_simulation_time);
	if (was_by_obstacle) rv = BUMPER_PUSHED;
	else rv = BUMPER_FREE;
	preprocess_simulation();
	debugmsg(DL_DEBUGGING, "---E--S1()-\n");
	return rv;
}    

/* converts the light value to 0..1023 scale */
int light_value(double l)
{
	int lv = (int)(l * 1023.0);
	if (lv > 1023) return 1023;
	else return lv;
}

/* upwards looking light sensor */
int SENSOR_2() 
{ 
	double light;
	int i, rv;

	debugmsg(DL_DEBUGGING, "---B--S2()-\n");
	update_simulation();
	debugmsg1long(DL_SIMULATOR, "[sim;%lu]:SENSOR_2()\n", new_simulation_time);
	// sum contribution from all lights
	light = 0.0;
	for (i = 0; i < lightcnt; i++)
	  light += light_intensity[i] /
	             ((light_x[i] - robot_x) * (light_x[i] - robot_x) +
	              (light_y[i] - robot_y) * (light_y[i] - robot_y));
	rv = light_value(light);
	preprocess_simulation();
	debugmsg(DL_DEBUGGING, "---E--S2()-\n");
	return rv;
}

/* downwards looking light sensor */
int SENSOR_3() 
{
	double light;
	int i, rv;

	debugmsg(DL_DEBUGGING, "---B--S3()-\n");
	update_simulation();
	debugmsg1long(DL_SIMULATOR, "[sim;%lu]:SENSOR_3()\n", new_simulation_time);
  	rv = NO_MARK_SENSOR_READING;
	for (i = 0; i < markcnt; i++)
		if ((robot_x >= mark_x[i]) && (robot_x <= mark_x[i] + mark_w[i]) && 
		    (robot_y >= mark_y[i]) && (robot_y <= mark_y[i] + mark_h[i]))
		    rv = mark_value[i];
	 preprocess_simulation();
	 debugmsg(DL_DEBUGGING, "---E--S3()-\n");
	 return rv;
}		

int LIGHT_1() { return LIGHT_1_val; }			// there is bumper on port1
int LIGHT_2() { return SENSOR_2(); }
int LIGHT_3() { return SENSOR_3(); }		

int ROTATION_1() { return ROTATION_1_val; }		// rotation sensors not supported
int ROTATION_2() { return ROTATION_2_val; }
int ROTATION_3() { return ROTATION_3_val; }

void ds_active(unsigned *sensor) {}				// assuming active mode on port2, otherwise passive bumpers
void ds_passive(unsigned *sensor) {}
void ds_rotation_set(unsigned *sensor, int pos) { *sensor = pos; }	// rotation sensors not supported
void ds_rotation_on(unsigned *sensor) {}
void ds_rotation_off(unsigned *sensor) {}
 
/*--------------------------------------------------------------------------------------
 motor handling (equivalent to dmotor.h, has to be updated to new version of LegOS 
 
 these are the functions that the controller uses to control the motors and they
 are equivalent to those in real LegOS
 
 */

MotorState dm_a,        //!< motor A state
  dm_b,                //!< motor B state
  dm_c;                //!< motor C state
	
//! set motor A direction
/*! \param dir the direction
 */
void motor_a_dir(MotorDirection dir)
{
  debugmsg(DL_DEBUGGING, "---B--MAdir()-\n");	
  update_simulation();
  debugmsg2long(DL_SIMULATOR, "[sim;%lu]:motor_a_dir(%ld)\n", new_simulation_time, (long)dir);
  dm_a.dir = dir;
  update_robot_speed();
  preprocess_simulation();
  debugmsg(DL_DEBUGGING, "---E--MAdir()-\n");	
}

//! set motor B direction			// no effect so far, should operate the highlift fork...
/*! \param dir the direction
 */
void motor_b_dir(MotorDirection dir)
{
  debugmsg(DL_DEBUGGING, "---B--MBdir()-\n");	
  update_simulation();
  debugmsg2long(DL_SIMULATOR, "[sim;%lu]:motor_b_dir(%ld)\n", new_simulation_time, (long)dir); 
  dm_b.dir = dir;
  preprocess_simulation();
  debugmsg(DL_DEBUGGING, "---E--MBdir()-\n");	
}

//! set motor C direction
/*! \param dir the direction
 */
void motor_c_dir(MotorDirection dir)
{
  debugmsg(DL_DEBUGGING, "---B--MCdir()-\n");	
  update_simulation();
  debugmsg2long(DL_SIMULATOR, "[sim;%lu]:motor_c_dir(%ld)\n", new_simulation_time, (long)dir);
  dm_c.dir = dir;
  update_robot_speed();
  preprocess_simulation();
  debugmsg(DL_DEBUGGING, "---E--MCdir()-\n");	
}

//! set motor A speed
/*! \param speed the speed
 */
void motor_a_speed(unsigned char speed)
{
  debugmsg(DL_DEBUGGING, "---B--MAspd()-\n");	
  update_simulation();
  debugmsg2long(DL_SIMULATOR, "[sim;%lu]:motor_a_speed(%ld)\n", new_simulation_time, (long)speed);
  dm_a.access.c.delta = speed;
  update_robot_speed();
  preprocess_simulation();
  debugmsg(DL_DEBUGGING, "---E--MAspd()-\n");	
}

//! set motor B speed  // this has no effect so far... only motors A and C... Should be the high lifting fork
/*! \param speed the speed
 */
void motor_b_speed(unsigned char speed)
{
  debugmsg(DL_DEBUGGING, "---B--MBspd()-\n");	
  update_simulation();
  debugmsg2long(DL_SIMULATOR, "[sim;%lu]:motor_b_speed(%ld)\n", new_simulation_time, (long)speed);
  dm_b.access.c.delta = speed;
  preprocess_simulation();
  debugmsg(DL_DEBUGGING, "---E--MBspd()-\n");	
}

//! set motor C speed
/*! \param speed the speed
 */
void motor_c_speed(unsigned char speed)
{
  debugmsg(DL_DEBUGGING, "---B--MCspd()-\n");	
  update_simulation();
  debugmsg2long(DL_SIMULATOR, "[sim;%lu]:motor_c_speed(%ld)\n", new_simulation_time, (long)speed);
  dm_c.access.c.delta = speed;
  update_robot_speed();
  preprocess_simulation();
  debugmsg(DL_DEBUGGING, "---B--MCspd()-\n");	
}

/* recomputes the robot speed based on the motor speed settings */
void update_robot_speed()
{
	int s1, s2;
	
    if (dm_a.dir == fwd) s1 = dm_a.access.c.delta;
    else if (dm_a.dir == rev) s1 = - dm_a.access.c.delta;
    else s1 = 0;
    if (dm_c.dir == fwd) s2 = dm_c.access.c.delta;
    else if (dm_c.dir == rev) s2 = - dm_c.access.c.delta;
    else s2 = 0;
    
   	robot_speed = robot_speed_ratio * 0.5 * (s1 + s2);
	if (robot_speed == 0)
		if (s1 == 0)
			robot_moving = MOVING_STOPPED;
		else if (s1 > s2)
			robot_moving = MOVING_ROTATING_RIGHT;
		else
			robot_moving = MOVING_ROTATING_LEFT;
	else if (robot_speed > 0)
		if (s1 == s2)
			robot_moving = MOVING_FWD_STRAIGHT;
		else if (s1 > s2)
		{
			robot_moving = MOVING_FWD_RIGHT;

			robot_c_r = robot_radius * (s1 + s2) / (s1 - s2);
			robot_c_x = robot_x + robot_c_r * cos(robot_heading);
			robot_c_y = robot_y - robot_c_r * sin(robot_heading); 
		}
		else
		{
			robot_moving = MOVING_FWD_LEFT;

			robot_c_r = robot_radius * (s1 + s2) / (s2 - s1);
			robot_c_x = robot_x + robot_c_r * cos(robot_heading);
			robot_c_y = robot_y - robot_c_r * sin(robot_heading); 
		}
	else  // robot speed < 0
		if (s1 == s2)
			robot_moving = MOVING_BWD_STRAIGHT;
		else if (s1 > s2)
		{
			robot_moving = MOVING_BWD_LEFT;

			robot_c_r = robot_radius * (s1 + s2) / (s2 - s1);
			robot_c_x = robot_x - robot_c_r * cos(robot_heading);
			robot_c_y = robot_y + robot_c_r * sin(robot_heading);
		}
		else 
		{
			robot_moving = MOVING_BWD_RIGHT;

			robot_c_r = robot_radius * (s1 + s2) / (s1 - s2);
			robot_c_x = robot_x - robot_c_r * cos(robot_heading);
			robot_c_y = robot_y + robot_c_r * sin(robot_heading);
		}
	debugmsg1double(DL_DEBUGGING, "robot speed updated to: %lf\n", robot_speed);
}

/*--------------------------------------------------------------------------------------
 other essential LegOS functions */

void msleep(int millisec)
{
    struct timespec t;

    t.tv_sec = millisec / 1000;
    t.tv_nsec = ((long)(millisec % 1000)) * 1000000;
    nanosleep(&t, NULL);
}

wakeup_t wait_event(wakeup_t(*wakeup) (wakeup_t), wakeup_t data)
{
  wakeup_t res;

  while (!(res = wakeup(data))) msleep(20);
  
  return res;
}

static sem_t exec_sem;        // coordinate starting of a new thread
static int (*process_start) (int, char **);
static int argc;

void *start_thread(void *arg)
{
    int argc_launch = argc;                // copy it from data segment to stack, so that we can unlock semaphore... :((  
    int (*thread_main) (int, char **) = process_start;

    sem_post(&exec_sem);    
    thread_main(argc_launch, (char **) arg);    // 3,2,1... go!
    return NULL;
}

pid_t execi(int (*code_start) (int, char **), int argc, char **argv,
           priority_t priority, size_t stack_size)
{
    pthread_t t;
    int err;

    sem_wait(&exec_sem);    // prevent more threads calling execi at the same time and cause conflicts

    process_start = code_start;
    if ((err = pthread_create(&t, NULL, start_thread, (void *) argv)) != 0)
    {
        perror("pthread_create failed");
        printf("error=%d\n", err);
        exit(1);
    }
    return t;
}

void simulated_exit(int code)        // code ignored.
{
    pthread_exit(NULL);
}

void yield(void)    // POSIX threads don't implement this :(((
{
    msleep(5);
}

time_t gettm()
{
    struct timeb tb;

    ftime(&tb);
    return tb.time * 1000 + tb.millitm;
}

/*--------------------------------------------------------------------------------------
 simulator types, variables, and functions */

/* noise functions */

/* speed of the robot is noisified after it is changed by the controller */
double noise_speed(double v)
{
	return v;
}

/* heading of the robot is noisified after it is changed by the controller */
double noise_heading(double d, double alpha)
{
	return alpha;
}

/* when robot hits wall or obstacle, its heading is noisified based on the distance the robot travelled
 * since last update */
double noise_heading_obstacle(double d, double alpha)
{
	return alpha;
}

/* displays information about the current project 
 * NOTE: when new types of obstacles, marks, etc. are introduced, the code has
 * to be modified to take them into accound (as well as init_simulator) */
void print_project_info(char *fileName)
{
	int i, j;
	
	debugmsg(DL_AUX, "-------------------------------------------------------------------------------\n");
	debugmsg(DL_AUX, "Project file: ");
	debugmsg(DL_AUX, fileName);
	debugmsg1int(DL_AUX, "\nEnvironment type: %d\n", envtype);
	debugmsg2double(DL_AUX, " width: %lf  height: %lf\n", env_w, env_h);
	debugmsg1int(DL_AUX, "#of obstacles: %d\n", obstcnt);
	for (i = 0; i < obstcnt; i++)
	{
		debugmsg2double(DL_AUX, " {%lf, %lf, ", obs_x[i], obs_y[i]);
		debugmsg2double(DL_AUX, "%lf, %lf}\n", obs_w[i], obs_h[i]);
	}
	debugmsg1int(DL_AUX, "#of floor marks: %d\n", markcnt);
	for (i = 0; i < markcnt; i++)
	{
		debugmsg2double(DL_AUX, " {%lf, %lf, ", mark_x[i], mark_y[i]);
		debugmsg2double(DL_AUX, "%lf, %lf, ", mark_w[i], mark_h[i]);
		debugmsg1int(DL_AUX, "%d}\n", mark_value[i]);
	}
    debugmsg1int(DL_AUX, "#of light sources: %d\n", lightcnt);
	for (i = 0; i < lightcnt; i++)
	{
		debugmsg2double(DL_AUX, " {%lf, %lf, ", light_x[i], light_y[i]);
		debugmsg1double(DL_AUX, "%lf}\n", light_intensity[i]);
	}
    debugmsg1int(DL_AUX, "#of active areas: %d\n", activecnt);
    for (i = 0; i < activecnt; i++)
    {
    	debugmsg1int(DL_AUX, " {%d, ", active_activation[i]);
    	if (active_activation[i] == ACTIVE_WITH_DELAY)
    	   debugmsg1int(DL_AUX, "[%d], ", active_delay[i]);
    	debugmsg2double(DL_AUX, "%lf, %lf, ", active_x[i], active_y[i]);
    	debugmsg2double(DL_AUX, "%lf, %lf, ", active_w[i], active_h[i]);
    	debugmsg1int(DL_AUX, "%d}\n", active_script_event[i]);
    }
    debugmsg1int(DL_AUX, "#of global time events: %d\n", eventcnt);
	for (i = 0; i < eventcnt; i++)
	{
		debugmsg1int(DL_AUX, " {%d, ", event_periodic[i]);
		if (event_periodic[i])
			debugmsg2long(DL_AUX, "(%ld, %ld), ", event_start[i], (long) event_count[i]);
		debugmsg2long(DL_AUX, "%ld, %ld}\n", event_time[i], event_script_event[i]);
	}
	debugmsg1int(DL_AUX, "#of event scripts: %d\n", scriptcnt);
	for (i = 0; i < scriptcnt; i++)
	{
		debugmsg(DL_AUX, " {");
		j = 0;
		while (script[i][j][0] != SCRIPT_STOP)
		{
			if (script[i][j][0] >= 10)
				debugmsg2int(DL_AUX, "(%d, %d) ", script[i][j][0], script[i][j][1]);
			else
				debugmsg1int(DL_AUX, "(%d) ", script[i][j][0]);
			j++;
		}
		debugmsg(DL_AUX, "}\n");
	}
	
	debugmsg1double(DL_AUX, "robot radius: %lf\n", robot_radius);
	debugmsg2double(DL_AUX, "initial location: %lf, %lf\n", init_x, init_y);
	debugmsg1double(DL_AUX, "initial heading: %lf\n", init_heading);
	debugmsg1double(DL_AUX, "slowdown constant: %lf\n", slowdown_constant);
	debugmsg(DL_AUX, "trajectory file : ");
	debugmsg(DL_AUX, trajectory_file);
	debugmsg(DL_AUX, "\n-------------------------------------------------------------------------------\n");
}

/* auxiliary function for processing input - ignoring the comment lines */
void readline(FILE *f, char *ln)
{
    readline_eof = 0;
    do {
      fgets(ln, 255, f);
    } while ((ln[0] == '#') && (!feof(f)));
    if (feof(f)) readline_eof = 1;
}

/* initialize simulator from a given file, if viewer is nonzero, simulation will
 * be shown in a window */
void init_simulator(char *fileName, int viewer)
{
    FILE *f;
    char ln[1000];
    int i, j, typeID;

    debugmsg(DL_INFO, "initializing the simulator...\n");

	viewer_on = viewer;
    
    f = fopen(fileName, "r");
    if (f == NULL) {
        printf("cannot open project file %s\n", fileName);
        exit(1);
    }
    
    readline(f, ln);        /* environment type */
    sscanf(ln, "%d", &envtype);
    
    if (envtype == ENVTYPE_RECTANGLE)
    {
        readline(f, ln);
        sscanf(ln, "%lf%lf", &env_w, &env_h);    /* environment dimensions */
    }
    
    /* read obstacle descriptions */
    readline(f, ln);
    sscanf(ln, "%d", &obstcnt);
    
    if (obstcnt > MAX_OBSTCNT) 
    {
        printf("Maximum # of obstacles exceeded. Modify a constant in simulator.h and recompile.\n");
        fclose(f);
        exit(0);
    }
    
    for (i = 0; i < obstcnt; i++)
    {
        readline(f, ln);
        sscanf(ln, "%d", &typeID);
        if (typeID == OBSTACLE_RECTANGLE)
        {
            sscanf(ln, "%*d %lf %lf %lf %lf", &obs_x[i], &obs_y[i], &obs_w[i], &obs_h[i]);
        }
    }
    
    /* read floor marks description */
    readline(f, ln);
    sscanf(ln, "%d", &markcnt);
    
    if (markcnt > MAX_MARKCNT)
    {
        printf("Maximum # of floor marks exceeded. Modify a constant in simulator.h and recompile.\n");
        fclose(f);
        exit(0);
    }
    
    for (i = 0; i < markcnt; i++)
    {
        readline(f, ln);
        sscanf(ln, "%d", &typeID);
        if (typeID == MARK_RECTANGLE)
        {
            sscanf(ln, "%*d %lf %lf %lf %lf", &mark_x[i], &mark_y[i], &mark_w[i], &mark_h[i], &mark_value[i]);
        }
    }
    
    /* read light sources description */
    readline(f, ln);
    sscanf(ln, "%d", &lightcnt);
    
    if (lightcnt > MAX_LIGHTCNT)
    {
        printf("Maximum # of light sources exceeded. Modify a constant in simulator.h and recompile.\n");
        fclose(f);
        exit(0);
    }
    
    for (i = 0; i < lightcnt; i++)
    {
        readline(f, ln);
        sscanf(ln, "%d", &typeID);
        if (typeID == LIGHT_POINTLIGHT)
        {
            sscanf(ln, "%*d %lf %lf %lf", &light_x[i], &light_y[i], &light_intensity[i]);
        }
    }
    
    /* read active areas description */
    readline(f, ln);
    sscanf(ln, "%d", &activecnt);
    
    if (activecnt > MAX_ACTIVECNT)
    {
        printf("Maximum # of active areas exceeded. Modify a constant in simulator.h and recompile.\n");
        fclose(f);
        exit(0);
    }
    
    for (i = 0; i < activecnt; i++)
    {
        readline(f, ln);
        sscanf(ln, "%d %d", &typeID, &active_activation[i]);
        if (typeID == ACTIVE_SCRIPT)
        	if (active_activation[i] == ACTIVE_WITH_DELAY)
        		sscanf(ln, "%*d %*d %d %lf %lf %lf %lf %d", &active_delay[i], &active_x[i], &active_y[i], &active_w[i], &active_h[i], &active_script_event[i]);
            else
            	sscanf(ln, "%*d %*d %lf %lf %lf %lf %d", &active_x[i], &active_y[i], &active_w[i], &active_h[i], &active_script_event[i]);
    }

    /* read global time events description */
    readline(f, ln);
    sscanf(ln, "%d", &eventcnt);

    if (eventcnt > MAX_EVENTCNT)
    {
    	printf("Maximum # of global time events exceeded. Modify a constant in simulator.h and recompile.\n");
    	fclose(f);
    	exit(0);
    }

	for (i = 0; i < eventcnt; i++)
	{
		readline(f, ln);
		sscanf(ln, "%d %d", &typeID, &event_periodic[i]);
		if (typeID == EVENT_SCRIPT)
		{
			if (event_periodic[i])
				sscanf(ln, "%*d %*d %ld %d %ld %d", &event_start[i], &event_count[i], &event_time[i], &event_script_event[i]);
			else
				sscanf(ln, "%*d %*d %ld %d", &event_time[i], &event_script_event[i]);
		}
	}

	/* read event scripts */
	readline(f, ln);
	sscanf(ln, "%d", &scriptcnt);
			
	for (i = 0; i < scriptcnt; i++)
	{
		j = 0;
        do {
        	if (j >= MAX_SCRIPT_SIZE)
            {
             	printf("Maximum script size exceeded. Active component #%d\n. Modify a constant in simulator.h and recompile.\n", i);
                fclose(f);
                exit(0);
            }
            readline(f, ln);
            sscanf(ln, "%d", &script[i][j][0]);
            if (script[i][j][0] >= 10)			// script command with argument?
            {
            	sscanf(ln, "%*d %d", &script[i][j][1]);
            }                
            j++;                
        } while (script[i][j-1][0] != SCRIPT_STOP);
    }
    
    /* type of robot */
    readline(f, ln);
    sscanf(ln, "%d", &typeID);
    
    /* robot dimensions */
    if (typeID == ROBOT_ROUND)
    {
        readline(f, ln);
        sscanf(ln, "%lf", &robot_radius);
    }

    /* robot speed ratio */
    readline(f, ln);
    sscanf(ln, "%lf", &robot_speed_ratio);
    
    /* initial location */
    readline(f, ln);
    sscanf(ln, "%lf %lf %lf", &init_x, &init_y, &init_heading);

    /* time slowdown constant */
    readline(f, ln);
    sscanf(ln, "%lf", &slowdown_constant);

	/* trajectory file */
	readline(f, ln);
	strcpy(trajectory_file, ln);
    
    fclose(f);    

    /* initialize the exec semaphore */
    sem_init(&exec_sem, 0, 1);

    print_project_info(fileName);
}

/* finalize simulator */
void close_simulator()
{
	FILE *f;
	int i;

	/* determine the final location of the robot */
	update_simulation();
	preprocess_simulation();
	
	/* close the simulation viewer */
	if (viewer_on) viewer_close();

	/* write the robot trajectory to a file */
	f = fopen(trajectory_file, "w+");
	fprintf(f, "moving\tx\ty\tc_r\tc_x\tc_y\tmoving\n");
	for (i = 0; i < traji; i++)
	{
		if ((traj_moving[i] != MOVING_FWD_RIGHT) && (traj_moving[i] != MOVING_FWD_LEFT) &&
		    (traj_moving[i] != MOVING_BWD_LEFT) && (traj_moving[i] != MOVING_BWD_RIGHT))
		{
			traj_c_r[i] = 0.0;
			traj_c_x[i] = 0.0;
			traj_c_y[i] = 0.0;
		}
		fprintf(f,"%d\t%lf\t%lf\t%lf\t%lf\t%lf\t%s\n", traj_moving[i], traj_x[i], traj_y[i],
			traj_c_r[i], traj_c_x[i], traj_c_y[i], moving_desc[traj_moving[i]]);
	}
	fclose(f);
}

/* execute single script command */
void execute_script_command(int cmd, int arg)
{
	switch (cmd)
	{
		case SCRIPT_LIGHTON:
			light_state[arg] = 1;
			break;
			
		case SCRIPT_LIGHTOFF:
		 	light_state[arg] = 0;
			break;

		case SCRIPT_REINITACTIVE:
			active_last_fired_at[arg] = 0;
			break;
	}
}

/* compute a distance of points x1,y1 and x2,y2 */
double dst(double x1, double y1, double x2, double y2)
{
	return sqrt((x1-x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
}

/* modifies the heading for the purpose of computation, or turns it back, if the robot is moving backwards */
void turn_if_backwards()
{
	if ((robot_moving == MOVING_BWD_STRAIGHT) ||
	    (robot_moving == MOVING_BWD_LEFT) ||
	    (robot_moving == MOVING_BWD_RIGHT))	
		robot_heading += pi;
}

/* finds the closest intersection of circle and rectangle, if the robot is standing at
 * location x0, y0 and moving either in clockwise (clockwise = 1) or anticlockwise (clockwise = 0)
 * direction. Rectangle is given by bottom left corner x,y, its width w, height h, and circle
 * is given by its radius c_r, and center c_x, c_y. Returns coordinates of intersection in ints_x, ints_y
 * and the returned value is the angle travelled on the circle, or > 2pi if no intersection */

/* used in aux. function - must be declared outside - "dirty C" */
static double dcr2;
static int clckws;
static double lna, lnb, lnc;
static double xi[8], yi[8], x00, y00;
static double alpha[8];

/* auxiliary function used in more formulas in the main function - "dirty C" */
void compute_alpha(int i)
{
	debugmsg1double(DL_DESPERATE, "1 - ((x00 - xi[0)^2) + (y00-yi[i])^2) / dcr2 = %lf\n", 1 - ((x00 - xi[i]) * (x00 - xi[i]) + (y00 - yi[i]) * (y00 - yi[i])) / dcr2);
	debugmsg2double(DL_DESPERATE, "x00 = %lf, y00 = %lf,", x00, y00);
	debugmsg2double(DL_DESPERATE, " xi[i] = %lf, yi[i] = %lf,", xi[i], yi[i]);
	debugmsg1double(DL_DESPERATE, " dcr2 = %lf\n", dcr2);

	alpha[i] = acos(1 - ((x00 - xi[i]) * (x00 - xi[i]) + (y00 - yi[i]) * (y00 - yi[i])) / dcr2);
	debugmsg1double(DL_DESPERATE, "alpha[i] eraly = %lf, ", alpha[i]);
	debugmsg2double(DL_DESPERATE, "lna = %lf, lnb = %lf, ", lna, lnb);
	debugmsg1double(DL_DESPERATE, "lnc = %lf\n", lnc);
	if (lna * xi[i] + lnb * yi[i] + lnc > 0)  // intersection lies in the first half circle going clockwise
		if (!clckws) alpha[i] = 2 * pi - alpha[i];
		else {}
	else
		if (clckws) alpha[i] = 2 * pi - alpha[i];
}

/* fn. description above */
double closest_intersection_of_circle_and_rectangle(int clockwise, double x0, double y0,
        double x, double y, double w, double h, 
        double c_x, double c_y, double c_r,
        double *ints_x, double *ints_y)
{
	double discr, bestalpha;
	int closest, i;

	debugmsg2double(DL_DEBUGGING, "  closest_intersection_of_circle_and_rectangle(%lf, %lf,\n", x0, y0);
	debugmsg2double(DL_DEBUGGING, "                                               %lf, %lf, ", x, y);
	debugmsg2double(DL_DEBUGGING, "%lf, %lf,\n", w, h);  	
	debugmsg2double(DL_DEBUGGING, "                                               %lf, %lf, ", c_x, c_y);
	debugmsg1double(DL_DEBUGGING, "%lf\n", c_r);
	
	// using the eq. of circle and boundaries of rectangle we find all the possible intersection
	// points (each side max. 2, i.e. max 8) and using cosine rule we obtain the angle to be 
	// travelled to reach each of these points from x0,y0 and return angle for the one that is closest
	
    dcr2 = 2 * c_r * c_r;	// auxiliary values used in multiple formulas below and above
	clckws = clockwise;
	// for the [x0,y0] - [x_c,y_c] line equation
	lna = y0 - c_y;
	lnb = c_x - x0;
	lnc = c_y * x0 - c_x * y0;
	x00 = x0;
	y00 = y0;
	    
	/* left side */
	discr = (c_r * c_r - (x  - c_x) * (x - c_x));	
	if (discr >= 0.0)
	{
		discr = sqrt(discr);
		xi[0] = x;
		yi[0] = c_y + discr;
		if ((yi[0] >= y) && (yi[0] <= y + h)) compute_alpha(0);	
		else alpha[0] = BEHIND_THE_HILL;

		xi[1] = x;
		yi[1] = c_y - discr;
		if ((yi[1] >= y) && (yi[1] <= y + h)) compute_alpha(1); 	
		else alpha[1] = BEHIND_THE_HILL;
	}
	else alpha[0] = alpha[1] = BEHIND_THE_HILL;
	debugmsg2double(DL_DESPERATE, "   alpha0,alpha1(left): %lf %lf\n", alpha[0], alpha[1]);
	
	/* right side */
	discr = (c_r * c_r - (x + w  - c_x) * (x + w - c_x));
	if (discr >= 0.0)
	{
		discr = sqrt(discr);
		xi[2] = x + w;
		yi[2] = c_y + discr;
		if ((yi[2] >= y) && (yi[2] <= y + h)) compute_alpha(2); 	
		else alpha[2] = BEHIND_THE_HILL;

		xi[3] = x + w;
		yi[3] = c_y - discr;
		if ((yi[3] >= y) && (yi[3] <= y + h)) compute_alpha(3);
		else alpha[3] = BEHIND_THE_HILL;
	}
	else alpha[2] = alpha[3] = BEHIND_THE_HILL;
	debugmsg2double(DL_DESPERATE, "   alpha2,alpha3(right): %lf %lf\n", alpha[2], alpha[3]);
	
	/* top side */
	discr = (c_r * c_r - (y + h - c_y) * (y + h - c_y));
	if (discr >= 0.0)
	{
		discr = sqrt(discr);
		xi[4] = c_x + discr;
		yi[4] = y + h;
		if ((xi[4] >= x) && (xi[4] <= x + w)) compute_alpha(4);
		else alpha[4] = BEHIND_THE_HILL;
	
		xi[5] = c_x - discr;
		yi[5] = y + h;
		if ((xi[5] >= x) && (xi[5] <= x + w)) compute_alpha(5);
		else alpha[5] = BEHIND_THE_HILL;
	}
	else alpha[4] = alpha[5] = BEHIND_THE_HILL;
	debugmsg2double(DL_DESPERATE, "   alpha4,alpha5(top): %lf %lf\n", alpha[4], alpha[5]);
	
	/* bottom side */
	discr = (c_r * c_r - (y - c_y) * (y - c_y));
	if (discr >= 0.0)
	{
		discr = sqrt(discr);
		xi[6] = c_x + discr;
		yi[6] = y;
		debugmsg2double(DL_DESPERATE, "   xi6 = %lf, yi6 = %lf\n", xi[6], yi[6]);
		
		if ((xi[6] >= x) && (xi[6] <= x + w)) compute_alpha(6);
		else alpha[6] = BEHIND_THE_HILL;

		xi[7] = c_x - discr;
		yi[7] = y;
		debugmsg2double(DL_DESPERATE, "   xi7 = %lf, yi7 = %lf\n", xi[7], yi[7]);

		if ((xi[7] >= x) && (xi[7] <= x + w)) compute_alpha(7);
		else alpha[7] = BEHIND_THE_HILL;
	}
	else alpha[6] = alpha[7] = BEHIND_THE_HILL;
	debugmsg2double(DL_DESPERATE, "   alpha6,alpha7(bottom): %lf %lf\n", alpha[6], alpha[7]);
	
	/* find the closest */
	bestalpha = BEHIND_THE_HILL;
	closest = -1;
	for(i = 0; i < 8; i++)
		if (alpha[i] < bestalpha)
		{
			bestalpha = alpha[i];
			closest = i;
		}
	if (closest >= 0)
	{
		*ints_x = xi[closest];
		*ints_y = yi[closest];
		debugmsg1int(DL_DESPERATE, "     -> closest: %d ", closest);
		debugmsg2double(DL_DESPERATE, "(%lf,%lf)\n", *ints_x, *ints_y);
		return alpha[closest];
	}
	else return BEHIND_THE_HILL;		// no intersection
}

/* determines the intersection point (sets into *ints_x, *ints_y) of robot trajectory
 * with a given rectangle based on its current location and heading, 
 * returns the distance to the intersection point or -1 if the trajectory doesn't
 * cross the rectangle. If the robot trajectory just touches the rectangle, 0 is returned.
 * Takes the dimensions of the robot into account, i.e. the intersection point is the
 * point of the robot center (or other reference point) where the robot body will touch 
 * the rectangle. 
 * inside argument specifies that the robot is inside the rectangle. if inside=0 and robot
 *        is still inside the triangle, its current position and 0 distance is returned.
 *        intersection with obstacles is called as inside = 0 and with environment edges
 *        as inside = 1 */
double robot_path_intersects(int inside, double x, double y, double w, double h, double *ints_x, double *ints_y)
{
	double r_r, bot_ints_x, top_ints_x, left_ints_y, right_ints_y;
	double beta;
	double xx, yy, ww, hh;
	int clcws;

	if ((robot_moving == MOVING_STOPPED) || 
	    (robot_moving == MOVING_ROTATING_LEFT) || 
	    (robot_moving == MOVING_ROTATING_RIGHT)) 
	    	return -1;		/* stopped or rotating -> no intersection */

	// BUG: we assume that the robot has a square shape, i.e. it will touch rectangle when
	// passing close to a corner in a way that a circular robot could pass, but square
	// robot coudn't.  (otherwise the computation would get a bit more complex...)

	// first a special case: robot is already inside the rectangle 
	if ((robot_x + robot_radius >= x) && (robot_x - robot_radius <= x + w) && 
	    (robot_y + robot_radius >= y) && (robot_y - robot_radius <= y + h) && !inside)
	{
		*ints_x = robot_x;
		*ints_y = robot_y;
		return 0.0;
	}
	
	if ((robot_moving == MOVING_FWD_STRAIGHT) || (robot_moving == MOVING_BWD_STRAIGHT))
	{
		turn_if_backwards();	/* moving straight. if backwards, turn around for the */
								/* moment of computation */
								 
		if (inside)			/* if robot inside the rectangle, "turn it inside out"... */
		{
			yy = y + h;
			hh = -h;
			xx = x + w;
			ww = -w;
		}
		else
		{
			yy = y;
			hh = h;
			xx = x;
			ww = w;
		}
		
		/* special cases: robot heading is 0, 180, 90, 270 */
		if (abs(robot_heading - 0.0) < epsilon)		// case 0
			if ((robot_x + robot_radius >= x) && (robot_x - robot_radius <= x + w))	// robot in x interval of the rectangle
        		if (robot_y + robot_radius <= yy)	// robot below the rectangle
        		{
          			*ints_x = robot_x;
          			*ints_y = yy - robot_radius;
          			turn_if_backwards();
          			return *ints_y - robot_y;
				}
				else	// robot above or out of the rectangle -> not reached
				{
					turn_if_backwards();
					return -1;
				}
			else // not in x interval of rectangle -> won't be reached
			{
				turn_if_backwards();
		  		return -1;
		  	}
		else if (abs(robot_heading - RADIAN180DEG) < epsilon)	// case 180
			if ((robot_x + robot_radius >= x) && (robot_x - robot_radius <= x + w))	// robot in x interval of the rectangle
				if (robot_y - robot_radius >= yy + hh)	// robot above the rectangle
				{
					*ints_x = robot_x;
					*ints_y = yy + hh + robot_radius;
					turn_if_backwards();
					return robot_y - *ints_y;
				}
				else 	// robot below or out of the rectangle -> not reached
				{
					turn_if_backwards();
					return -1;
				}
			else	// not in x interval of rectangle -> won't be reached
			{
				turn_if_backwards();
				return -1;
			}
		else if (abs(robot_heading - RADIAN90DEG) < epsilon)	// case 90
			if ((robot_y + robot_radius >= y) && (robot_y - robot_radius <= y + h)) // robot in y interval of the rectangle
				if (robot_x + robot_radius <= xx)	// robot on the left of the rectangle
				{
					*ints_x = xx - robot_radius;
					*ints_y = robot_y;
					turn_if_backwards();
					return *ints_x - robot_x;
				}
				else 	// robot right to or out of the rectangle -> not reached
				{
					turn_if_backwards();
					return -1;
				}
			else	// not in y interval of the rectangle
			{
				turn_if_backwards();
				return -1; 
			}
		else if (abs(robot_heading - RADIAN270DEG) < epsilon)	// case 270
			if ((robot_y + robot_radius >= y) && (robot_y - robot_radius <= y + h)) // robot in y interval of the rectangle
				if (robot_x - robot_radius >= xx + ww)	// robot on the right of the rectangle
				{
					*ints_x = xx + ww + robot_radius;
					*ints_y = robot_y;
					turn_if_backwards();
					return robot_x - *ints_x;
				}
				else	// robot left to or out of the rectangle -> not reached
				{
					turn_if_backwards();
					return -1;
				}
			else	// not in y interval of the rectangle
			{
				turn_if_backwards();
				return -1;
			}
		else	// heading not parallel with axes (general case)
		{
			/* compute intersections with all 4 sides of rectangle (to improve efficiency, copy directly
			 * to quadrants below...) */

			if (robot_y <= yy)	bot_ints_x = (yy - robot_radius - robot_y) * tan(robot_heading) + robot_x;
			if (robot_y >= yy + hh) top_ints_x = (yy + hh - robot_radius - robot_y) * tan(robot_heading) + robot_x;
			if (robot_x < xx)	left_ints_y = (xx - robot_radius - robot_x) / tan(robot_heading) + robot_y;
			if (robot_x >= xx + ww) right_ints_y = (xx + ww + robot_radius - robot_x) / tan(robot_heading) + robot_y;

			/* now depending on the quadrant of the robot heading determine the intersection point and return
			 * the distance */
			
			if (robot_heading < RADIAN90DEG)  /* first quadrant 0 - 90 */
				if ((left_ints_y >= y) && (left_ints_y <= y + h))	/* left side? */
				{	
					*ints_x = xx;
					*ints_y = left_ints_y;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else if ((bot_ints_x >= x) && (bot_ints_x <= x + w)) /* bottom side? */
				{
					*ints_x = bot_ints_x;
					*ints_y = yy;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else	// neither
				{
					turn_if_backwards();
					return -1;
				}	
			else if (robot_heading < RADIAN180DEG)	/* second quadrant 90 - 180 */
				if ((left_ints_y >= y) && (left_ints_y <= y + h)) 	/* left side? */
				{
					*ints_x = xx;
					*ints_y = left_ints_y;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else if ((top_ints_x >= x) && (top_ints_x <= x + w)) /* top side? */
				{
					*ints_x = top_ints_x;
					*ints_y = yy + hh;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else	// neither
				{
					turn_if_backwards();
					return -1;
				}	
			else if (robot_heading < RADIAN270DEG) /* third quadrant 180 - 270 */
				if ((top_ints_x >= x) && (top_ints_x <= x + w)) /* top side? */
				{
					*ints_x = top_ints_x;
					*ints_y = yy + hh;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else if ((right_ints_y >= y) && (right_ints_y <= y + h)) /* right side? */
				{
					*ints_x = xx + ww;
					*ints_y = right_ints_y;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else	// neither
				{
					turn_if_backwards();
					return -1;
				}					
			else 	/* last quadrant 270 - 360 */
				if ((right_ints_y >= y) && (right_ints_y <= y + h)) /* right side? */
				{
					*ints_x = xx + ww;
					*ints_y = right_ints_y;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else if ((bot_ints_x >= x) && (bot_ints_x <= x + w)) /* bottom side? */
				{
					*ints_x = bot_ints_x;
					*ints_y = yy;
					turn_if_backwards();
					return dst(robot_x, robot_y, *ints_x, *ints_y);
				}
				else	// neither
				{
					turn_if_backwards();
					return -1;
				}
		}
	}
	else 		/* robot moving along a circle */
	{
		// check the intersection of the trajectory of robot reference point with the rectangle
		//  - enlarge the rectangle by robot radius so that we track also hits of the robot edges
		if ((robot_moving == MOVING_FWD_LEFT) || (robot_moving == MOVING_BWD_RIGHT)) 
		   clcws = 0;
		else clcws = 1;

		if (inside) r_r = - robot_radius;			// make the rectangle smaller instead of larger
		else r_r = robot_radius;
		
		beta = closest_intersection_of_circle_and_rectangle(clcws, robot_x, robot_y, 
			  x - r_r, y - r_r, w + 2 * r_r, h + 2 * r_r,
			  robot_c_x, robot_c_y, robot_c_r,
			  ints_x, ints_y);

		if (beta > 2 * pi)	/* no intersection */
			return -1;
		else
			return beta * robot_c_r;
	}
	
	return 0.0; 	/* never reached */
}
	
/* insert the local time event into the list at the proper position, i.e. keeping it sorted */
void insert_time_event(unsigned long time, int script_index, int active_index)
{
	int i, j;

	debugmsg2int(DL_DEBUGGING, "adding local time event, script index: %d, active index: %d, time: ", script_index, active_index);
	debugmsg1long(DL_DEBUGGING, "%lu\n", time);
	
	/* find the index to insert */
	for(i = 0; i < te_len; i++)
		if (time_events[i].time >= time) break;
	
	/* shift the rest of the list one position right */
	for(j = te_len; j--; j != i)
	{
		time_events[j].time 		= time_events[j - 1].time;
		time_events[j].script_index = time_events[j - 1].script_index;
		time_events[j].active_index = time_events[j - 1].active_index;
	}

	/* and save the new item */
	time_events[i].time = time;
	time_events[i].script_index = script_index;
	time_events[i].active_index = active_index;

	te_len++;
}

/* removes all events from local time event list */
inline void clear_time_events()
{
	te_len = 0;
} 

/* updates the time_events array based on current robot heading, speed, and location and
 * the active areas and obstacles */
void update_time_events()
{
	double ints_x = -1.0, ints_y = -1.0;
	double min_ints_d, ints_d;
	double bot_ints_x, top_ints_x, left_ints_y, right_ints_y;
	int closest;
	int i;
	
   /* for all obstacles that are in the way, find the one that is closest */

   if ((robot_moving != MOVING_STOPPED) && 
	   (robot_moving != MOVING_ROTATING_LEFT) && 
	   (robot_moving != MOVING_ROTATING_RIGHT))
   {
   		min_ints_d = env_w + env_h;
   		closest = -1;
   
   		for (i = 0; i < obstcnt; i++)
     	  if ((ints_d = robot_path_intersects(0, obs_x[i], obs_y[i], obs_w[i], obs_h[i], &ints_x, &ints_y)) >= 0)
	  	  {
	    	debugmsg1int(DL_DEBUGGING, "robot path intersects with obstacle #%d\n", i);
      		if (ints_d < min_ints_d)
       		{
       		  min_ints_d = ints_d;
       		  closest = i;
       		}
       	  }

   		if (closest == -1) /* no obstacle - take the wall instead */
		{
			debugmsg(DL_DEBUGGING, "robot path doesn't intersect with any obstacle, using wall\n");
   			min_ints_d = robot_path_intersects(1, 0.0, 0.0, env_w, env_h, &ints_x, &ints_y);
		}

		next_obstacle_x = ints_x;
		next_obstacle_y = ints_y;

		debugmsg2double(DL_DEBUGGING, "about to put the collision point on event list, min_inst_d = %lf, robot_speed = %lf\n", min_ints_d, robot_speed);		

		/* put the next collision point (obstacle or wall) on the time event list */
		if (min_ints_d > 0.0) insert_time_event(new_simulation_time + (long)(min_ints_d / robot_speed), -1, -1);
		else debugmsg(DL_DEBUGGING, "no, there is no intersection with wall neither\n");
	}

	/* now process all the active areas and in case of intersection, put them on the
	 * time event list */
	for (i = 0; i < activecnt; i++)
		if ((ints_d = robot_path_intersects(0, active_x[i], active_y[i], active_w[i], active_h[i], &ints_x, &ints_y)) >= 0)
			insert_time_event(new_simulation_time + (long)(ints_d / robot_speed), active_script_event[i], i);
}

/* this code was used before to compute intersection with the environment edges 
 * 	// compute intersections with all 4 sides of the wall:
			bot_ints_x = robot_x - (robot_y - robot_radius) * tan(robot_heading);
			top_ints_x = (env_h - robot_radius - robot_y) * tan(robot_heading) + robot_x;
			left_ints_y = robot_y - (robot_x - robot_radius) / tan(robot_heading);
			right_ints_y = (env_w - robot_radius - robot_x) / tan(robot_heading) + robot_y;
		
        // based on quadrant of robot heading, find the intersection point...
        if (robot_heading < RADIAN90DEG)  // first quadrant 0 - 90 
			if ((right_ints_y >= 0.0) && (right_ints_y <= env_h))	// right side? 
			{	
				ints_x = env_w;
				ints_y = right_ints_y;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with right side at %lf, %lf\n", ints_x, ints_y);
			}
			else		// must be the top side 
			{
				ints_x = top_ints_x;
				ints_y = env_h;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with top side at %lf, %lf\n", ints_x, ints_y);
			}
		else if (robot_heading < RADIAN180DEG)	// second quadrant 90 - 180 
			if ((right_ints_y >= 0.0) && (right_ints_y <= env_h)) 	// right side? 
			{
				ints_x = env_w;
				ints_y = right_ints_y;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with right side at %lf, %lf\n", ints_x, ints_y);
			}
			else 		// must be the bottom side 
			{
				ints_x = bot_ints_x;
				ints_y = 0.0;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with bottom side at %lf, %lf\n", ints_x, ints_y);
			}
		else if (robot_heading < RADIAN270DEG) // third quadrant 180 - 270 
			if ((bot_ints_x >= 0.0) && (bot_ints_x <= env_w)) // bottom side? 
			{
				ints_x = bot_ints_x;
				ints_y = 0.0;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with bottom side at %lf, %lf\n", ints_x, ints_y);
			}
			else	  // must be the left side 
			{
				ints_x = 0.0;
				ints_y = left_ints_y;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with left side at %lf, %lf\n", ints_x, ints_y);
			}
		else 	// last quadrant 270 - 360 
			if ((left_ints_y >= 0.0) && (left_ints_y <= env_h)) // left side? 
			{
				ints_x = 0.0;
				ints_y = left_ints_y;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with left side at %lf, %lf\n", ints_x, ints_y);

			}
			else 	 // must be the top side 
			{
				ints_x = top_ints_x;
				ints_y = env_h;
				min_ints_d = dst(robot_x, robot_y, ints_x, ints_y);
				debugmsg2double(DL_DEBUGGING, "intersection with top side at %lf, %lf\n", ints_x, ints_y);
			}

*/

/* start the simulation: resets the environment, robot and the controller,
   starts the global simulation time */
void start_simulator()
{
	int i, j, k;

	debugmsg(DL_INFO, "starting the simulator...\n"); 
	
    /* initialize robot location, heading, and speed */
	robot_moving = MOVING_STOPPED;
    robot_x = init_x;
    robot_y = init_y;
    robot_heading = last_robot_heading = init_heading;
    robot_speed = last_robot_speed = 0.0;

    /* initialize active areas */
    for(i = 0; i < activecnt; i++)
    	if (active_activation[i] == ACTIVE_WITH_DELAY)
    		active_last_fired_at[i] = - active_delay[i] - 1;
    	else
    		active_last_fired_at[i] = 0;
    
    /* initialize all lights */
    for (i = 0; i < lightcnt; i++)
    	light_state[i] = 0;

    /* execute all time events at simulation time 0, if any */
	for (i = 0; i < eventcnt; i++)
		if ((event_periodic[i] && (event_start[i] == 0) && (event_count[i] > 0)) ||
		    (!event_periodic[i] && (event_time[i] == 0)))
		{
			k = event_script_event[i];
		    for (j = 0; script[k][j][0] != 0; j++)
		    	execute_script_command(script[k][j][0], script[k][j][1]);
		}

	/* initialize local time events */
	clear_time_events();
		
	/* initiliaze statistics variables */
	num_hits = 0;
	total_distance = 0.0;
	was_by_obstacle = 0;

    /* start the simulation time */
    simulation_time = 0;
    system_time = gettm();

    /* initialize the time events and compute the next obstacle/wall collision point */
    update_time_events();

	/* initialize trajectory store */
	traj_x[0] = init_x;
	traj_y[0] = init_y;
	traj_moving[0] = MOVING_STOPPED;
	traji = 1;

	/* initialize viewer window */
    if (viewer_on) viewer_open(WINDOW_WIDTH, WINDOW_HEIGHT, env_w, env_h);
}

/* called each time the controller access the simulator before performing actual action to
 * update to the current state of the simulation - we don't run in the background of the controller */
void update_simulation()
{
	int location_updated = 0;
	int is_by_obstacle;
	int i, j, k, first, ok_to_activate;
	long time_movement_stopped;
	double dt, new_x, new_y;
	
	/* stop the simulation time, while updating the simulation */
	new_system_time = gettm();
	new_simulation_time = (long) ((new_system_time - system_time) * slowdown_constant + simulation_time);
	time_movement_stopped = new_simulation_time;   // if robot runs into obstacle or wall, this will be modified below

	print_simstat();

	/* put relevant (i.e. simulation_time .. new_simulation_time) global time events 
	 * on the list of local time events */
	for (i = 0; i < eventcnt; i++)
	{
		if (event_periodic[i])	
			if ((event_start[i] <= new_simulation_time) && 				// current simulation time interval overlaps with active period of this event?
				(event_start[i] + (event_count[i] - 1) * event_time[i] > simulation_time))
			{
				if (event_start[i] < simulation_time)
					first = simulation_time + event_time[i] - ((simulation_time - event_start[i]) % event_time[i]);
				else
					first = event_start[i];
				for (j = first; (j <= new_simulation_time) && (j < event_start[i] + event_count[i] * event_time[i]); j += event_time[i])
					insert_time_event(j, event_script_event[i], -1);
			}
			else {}
		else
			if ((event_time[i] > simulation_time) && (event_time[i] <= new_simulation_time))
				insert_time_event(event_time[i], event_script_event[i], -1);
	}

	if (time_events[0].time > simulation_time) can_keep_events = 1;
	else can_keep_events = 0;

    is_by_obstacle = 0;
	/* process local time events to the point of new_simulation_time */
	for(i = 0; i < te_len; i++)
	{
		// process events only til new_simulation_time
		if (time_events[i].time > new_simulation_time) break;
		debugmsg1long(DL_DEBUGGING, "local time event: %lu\n", time_events[i].time);
		// fetch the script index
		j = time_events[i].script_index;
	    /* obstacle/wall? */
		if (j == -1)
		{
	    	/* yes -> update position to be in front of the obstacle,  
	     	 * number of hits, set flag to not update position later */
			debugmsg(DL_DEBUGGING, "ran into obstacle, updating location\n");
	     	new_x = next_obstacle_x;
	     	new_y = next_obstacle_y;
			dt = dst(robot_x, robot_y, new_x, new_y);
			robot_heading = noise_heading_obstacle(dt, robot_heading);
			if (!was_by_obstacle) num_hits++;
			is_by_obstacle = 1;
			location_updated = 1;
			time_movement_stopped = time_events[i].time;
			break; // don't process other events, we are stucked in an obstacle/wall
		}
		else
		{	/* either active area or time event. if active area, check whether it can be activated */
			ok_to_activate = 1;
			if ((k = time_events[i].active_index) >= 0)	/* active area? */
			{ 
				if (active_activation[k] == ACTIVE_ONE_TIME_ONLY)
					if (active_last_fired_at[k])
						ok_to_activate = 0;			/* has been already activated */
					else
						active_last_fired_at[k] = 1;
				else if (active_activation[k] == ACTIVE_WITH_DELAY)
					if (time_events[i].time - active_delay[k] < active_last_fired_at[k])
						ok_to_activate = 0;			/* delay didn't pass yet */
					else 
						active_last_fired_at[k] = time_events[i].time;
			}

			if (ok_to_activate)
			{						/* script event that is relevant, execute the script... */
				k = 0;
				while (script[j][k][0] != SCRIPT_STOP)
				{
					execute_script_command(script[j][k][0], script[j][k][1]);
					k++;
				}
			}
		}
	}
	was_by_obstacle = is_by_obstacle;
	
	/* if position not updated yet, update it */
	if (!location_updated)
	{
		debugmsg(DL_DEBUGGING, "wall/obstacle not reached, updating location\n");
		/* compute the distance travelled and new estimated location */
		dt = (new_simulation_time - simulation_time) * robot_speed;

    	/* compute the new coordinates */
    	new_x = robot_x + dt * sin(robot_heading);
    	new_y = robot_y + dt * cos(robot_heading);
    }

    total_distance += dt;
    robot_x = new_x;
    robot_y = new_y;

    // update also robot's heading if moving along a circular trajectory or rotating

    if ((robot_moving == MOVING_FWD_RIGHT) || (robot_moving == MOVING_BWD_LEFT))
// was:    	robot_heading += (new_simulation_time - time_movement_stopped) * robot_speed / robot_c_r;
    	robot_heading += (time_movement_stopped - simulation_time) * robot_speed / robot_c_r;
    else if ((robot_moving == MOVING_FWD_LEFT) || (robot_moving == MOVING_BWD_RIGHT))
    	robot_heading -= (time_movement_stopped - simulation_time) * robot_speed / robot_c_r;
    else if (robot_moving == MOVING_ROTATING_RIGHT)
    	robot_heading += (time_movement_stopped - simulation_time) * dm_a.access.c.delta / robot_radius;
    else if (robot_moving == MOVING_ROTATING_LEFT)
    	robot_heading -= (time_movement_stopped - simulation_time) * dm_a.access.c.delta / robot_radius;

	print_simstat();
	
	/* return to execute the operation */
}

/* prints the simulation status to screen and file, if enabled */
print_simstat()
{
	debugmsg2double(DL_SIMSTATUS, "SIMSTAT: robot: [%lf,%lf] ", robot_x, robot_y);
	debugmsg2double(DL_SIMSTATUS, "(%lf) speed: %lf ", robot_heading, robot_speed);
	debugmsg(DL_SIMSTATUS, moving_desc[robot_moving]);
	if ((robot_moving == MOVING_FWD_LEFT) || (robot_moving == MOVING_FWD_RIGHT) ||
	    (robot_moving == MOVING_BWD_LEFT) || (robot_moving == MOVING_BWD_RIGHT))
	{
	    debugmsg2double(DL_SIMSTATUS, "([%lf,%lf],", robot_c_x, robot_c_y);
	    debugmsg1double(DL_SIMSTATUS, "%lf)\n", robot_c_r);
	}
	else debugmsg(DL_SIMSTATUS, "\n");
	debugmsg2int(DL_SIMSTATUS, "         motors: (%d, %d), ", dm_a.dir, dm_a.access.c.delta);
	debugmsg2int(DL_SIMSTATUS, "(%d, %d)\n", dm_c.dir, dm_c.access.c.delta);
	debugmsg2long(DL_SIMSTATUS, "         simulation time: %ld, since last call: %ld\n", new_simulation_time, new_simulation_time - simulation_time);
}

/* updates the log of robot's trajectory */
void update_trajectory()
{
	if ((robot_moving != MOVING_STOPPED) && (robot_moving != MOVING_ROTATING_LEFT) &&
		(robot_moving != MOVING_ROTATING_LEFT))
	{
		if ((traj_x[traji - 1] == robot_x) &&
		    (traj_y[traji - 1] == robot_y)) traji--;		/* if position didn't change, erase previous point */
		traj_moving[traji] = robot_moving;
		traj_x[traji] = robot_x;
		traj_y[traji] = robot_y;
		traj_c_r[traji] = robot_c_r;
		traj_c_x[traji] = robot_c_x;
		traj_c_y[traji] = robot_c_y;
		if (traji < MAX_TRAJECTORY_LEN) traji++;
	}
}

/* called each time the controller accesses the simulator after performing actual action to
 * update the time events array, new obstacle position, and update the simulation and system time */
void preprocess_simulation()
{
	if (!((can_keep_events) && (last_robot_moving == robot_moving) && 
	     (last_robot_speed == robot_speed) &&
	     (last_robot_c_r == robot_c_r) &&
	     (last_robot_c_x == robot_c_x) &&
	     (last_robot_c_y == robot_c_y)))
	{
		// reinitialize the local time event list
		clear_time_events();
		update_time_events();

		last_robot_c_r = robot_c_r;
		last_robot_c_x = robot_c_x;
		last_robot_c_y = robot_c_y;
		last_robot_speed = robot_speed;

		update_trajectory();
	}

	/* update simulator viewer */	
	if (viewer_on) viewer_refresh();
	
    /* finally, start the time again */
	simulation_time = new_simulation_time;
	system_time = gettm();
}
