/* simulator.h   Type definitions and constants related to the simulator */

#ifndef    __simulator_h__
#define __simulator_h__

#include <time.h>
#include <pthread.h>

//! the motor directions
typedef enum {
  off = 0,            //!< freewheel
  fwd = 1,            //!< forward
  rev = 2,            //!< reverse
  brake = 3            //!< hold current position

} MotorDirection;

//! the motor status type.
typedef struct {
  union {
    unsigned assembler;        //!< assures word alignment for assembler

    struct {
      unsigned char delta;    //!< the speed setting
      unsigned char sum;    //!< running sum
    } c;
  } access;            //!< provides access from C and assembler
  MotorDirection dir;        //!< output pattern when sum overflows
} MotorState;

// legOS system types
typedef unsigned char wakeup_t;
typedef int priority_t;

#define  MIN_SPEED    0         //!< minimum motor speed        
#define  MAX_SPEED    255       //!< maximum motor speed

#define PRIO_NORMAL            0    // not used in simulator
#define DEFAULT_STACK_SIZE    0    // not used in simulator

#define WINDOW_WIDTH	320		// simulator viewer window dimensions
#define WINDOW_HEIGHT	300

/* these are the constants used in project file */

#define ENVTYPE_RECTANGLE    1
#define ENVTYPE_TORUS        2

#define OBSTACLE_RECTANGLE    1
#define MARK_RECTANGLE        1
#define LIGHT_POINTLIGHT      1
#define ACTIVE_SCRIPT        1
#define ACTIVE_ONE_TIME_ONLY	1
#define ACTIVE_ON_EACH_ENTRANCE	2
#define ACTIVE_WITH_DELAY		3
#define EVENT_SCRIPT			1
#define ROBOT_ROUND            1

// script commands without argument have id < 10
// otherwise 1 integer argument follows
#define SCRIPT_STOP        0
#define SCRIPT_LIGHTON        10
#define SCRIPT_LIGHTOFF        11
#define SCRIPT_REINITACTIVE		12

#define MAX_OBSTCNT         50
#define MAX_MARKCNT         50
#define MAX_LIGHTCNT        50
#define MAX_ACTIVECNT       10
#define MAX_EVENTCNT		10		// global time events
#define MAX_EVENT_SCRIPTS	10
#define MAX_SCRIPT_SIZE     20
#define MAX_LOCALTIMEEVENTS	40		// local time events

#define BUMPER_FREE				100		/* sensor value for free bumper */
#define BUMPER_PUSHED			800		/* sensor value for pushed bumper */

#define NO_MARK_SENSOR_READING	1000	/* sensor reading returned when there is no floor mark below the robot */

/* constants for viewer argument of init_simulator() */
#define VIEWER_ON	1	
#define VIEWER_OFF	0

#define MAX_TRAJECTORY_LEN	1000		/* of how many verteces can the trajectory consist */
	
/* function prototypes */

/* legOS functions */

#define exit        simulated_exit
#define sys_time    gettm()
#define pid_t        pthread_t

int SENSOR_1(); 
int SENSOR_2(); 
int SENSOR_3();

int LIGHT_1();
int LIGHT_2();
int LIGHT_3();

int ROTATION_1();
int ROTATION_2();
int ROTATION_3();

void ds_active(unsigned *sensor);
void ds_passive(unsigned *sensor);
void ds_rotation_set(unsigned *sensor, int pos);
void ds_rotation_on(unsigned *sensor);
void ds_rotation_off(unsigned *sensor);
 
void motor_a_dir(MotorDirection dir);
void motor_b_dir(MotorDirection dir);
void motor_c_dir(MotorDirection dir);
void motor_a_speed(unsigned char speed);
void motor_b_speed(unsigned char speed);
void motor_c_speed(unsigned char speed);

void msleep(int millisec);
wakeup_t wait_event(wakeup_t(*wakeup) (wakeup_t), wakeup_t data);
pid_t execi(int (*code_start) (int, char **), int argc, char **argv,
           priority_t priority, size_t stack_size);
void exit(int code);
void yield(void);
time_t gettm();

/* simulator functions */

void init_simulator(char *fileName, int viewer);
void start_simulator();
void close_simulator();

#endif //__simulator_h__