/* testrun.c    tests the run of the simulator with modules. Simply tests the modules
                NAVIGATE, CARGO, LINEFOLLOWER, MOTORDRIVER and AVOIDANCE.

                The evolutionary algorithm might use similar way how to test the 
                evolved program.
                
                Reads the simulator configuration from testrun.cfg file.

*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "modules.h"
#include "simulator.h"
#include "viewer.h"
#include "messagepassing.h"
#include "debug.h"

//#define PROJECT_FILE "H:\\rcx\\s\\cargo\\early\\testrun.prj"
#define PROJECT_FILE "testrun.prj"


int main(int argc, char **argv)
{
    unsigned char data[32];
    unsigned char done;
    unsigned char **usptr;

    initdebug(DL_INFO, DL_ALL - DL_DESPERATE);

	// initialize message passing and simulator
    mp_init();
	init_simulator(PROJECT_FILE, VIEWER_ON);

    debugmsg(DL_INFO, "registering and initiliazing the modules...\n");

    // register modules for message passing and initialize them
    mid_console = mp_register(console_message, IMMEDIATE_MODE);
    mid_navigate = mp_register(navigate_message, IMMEDIATE_MODE);
    mid_cargo = mp_register(cargo_message, QUEUE_MODE);
    mid_motordrv = mp_register(motordriver_message, IMMEDIATE_MODE);
    mid_lnflwer = mp_register(lnflwer_message, QUEUE_MODE);
    mid_avoidance = mp_register(avoidance_message, QUEUE_MODE);

    debugmsg2int(DL_AUX, "%d:console\n%d:navigate\n", mid_console, mid_navigate);
    debugmsg2int(DL_AUX, "%d:cargo\n%d:motordrv\n", mid_cargo, mid_motordrv);
    debugmsg2int(DL_AUX, "%d:lnflwer\n%d:avoidance\n\n",mid_lnflwer,mid_avoidance);

    start_simulator();

    debugmsg(DL_INFO, "**** testing module NAVIGATE...\n");

    debugmsg(DL_INFO, "NAVIGATE_FWD 10...\n");
    data[NAVIGATE_N] = 10;
    *((unsigned char **)(&data[NAVIGATE_FINISHED])) = &done;    
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_FWD, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(20);

    debugmsg(DL_INFO, "NAVIGATE_RIGHT 10...\n");    
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_RIGHT, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(20);
    
    debugmsg(DL_INFO, "NAVIGATE_LEFT 20...\n");
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_LEFT, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(20);
    
    debugmsg(DL_INFO, "NAVIGATE_TURNRND...\n");
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_TURNRND, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(20);
    
    debugmsg(DL_INFO, "NAVIGATE_FWD 15, NAVIGATE_STOP...\n");        
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_FWD, NAVIGATE_MSGSIZE, data);
    msleep(300);
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_STOP, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(20);

    debugmsg(DL_INFO, "NAVIGATE_SLOW, NAVIGATE_FWD 10, NAVIGATE_FAST, NAVIGATE_FWD 10...\n");
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_SLOW, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(5);
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_FWD, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(20);    
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_FAST, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(5);
    mp_send(MSG_FROMSYSTEM, mid_navigate, NAVIGATE_FWD, NAVIGATE_MSGSIZE, data);
    while (!done) msleep(20);
        
    debugmsg(DL_INFO, "**** testing module CARGO...\n");
    
    debugmsg(DL_INFO, "CARGO_LOAD...\n");
    mp_send(MSG_FROMSYSTEM, mid_cargo, CARGO_LOAD, 0, NULL);
    msleep(1000);

    debugmsg(DL_INFO, "CARGO_UNLOAD...\n");
    mp_send(MSG_FROMSYSTEM, mid_cargo, CARGO_UNLOAD, 0, NULL);
    msleep(1000);

    debugmsg(DL_INFO, "**** testing module LINEFOLLOWER...\n");
    
    data[0] = 10;
    debugmsg(DL_INFO, "FOLLOW 10...\n");
    mp_send(MSG_FROMSYSTEM, mid_lnflwer, LNFLWER_FOLLOW, 1, data);
    msleep(1000);

    debugmsg(DL_INFO, "**** testing module MOTORDRIVER...\n");

    debugmsg(DL_INFO, "MOTORDRV_POWER ABC 127...\n");
    data[0] = 7;
    data[1] = 127;
    mp_send(MSG_FROMSYSTEM, mid_motordrv, MOTORDRV_POWER, 2, data);
    msleep(1000);

    debugmsg(DL_INFO, "MOTORDRV_POWER ABC -127...\n");
    data[1] = (unsigned char)(-127);
    mp_send(MSG_FROMSYSTEM, mid_motordrv, MOTORDRV_POWER, 2, data);
    msleep(1000);

    debugmsg(DL_INFO, "**** testing module AVOIDANCE...\n");
    
    debugmsg(DL_INFO, "AVOIDANCE_START...\n");
    mp_send(MSG_FROMSYSTEM, mid_avoidance, AVOIDANCE_START, 0, NULL);
    msleep(5000);

    debugmsg(DL_INFO, "...end of tests, terminating modules...\n");
    
    debugmsg(DL_INFO, "navigate...\n");
    mp_goodbye(mid_navigate);
    debugmsg(DL_INFO, "cargo...\n");
    mp_goodbye(mid_cargo);
    debugmsg(DL_INFO, "motordrv...\n");
    mp_goodbye(mid_motordrv);
    debugmsg(DL_INFO, "lnflwer...\n");
    mp_goodbye(mid_lnflwer);
    debugmsg(DL_INFO, "avoidance...\n");
    mp_goodbye(mid_avoidance);
    
    msleep(1000);    
    debugmsg(DL_INFO, "console...\n");
    mp_goodbye(mid_console);

	debugmsg(DL_INFO, "closing simulator...\n");
    close_simulator();
    
    debugmsg(DL_INFO, "...bye\n");
    return 0;
}
