/* viewer.c   shows the progress of simulation in the graphical window   
*/

#include <g2.h>
#include <g2_X11.h>
#include <math.h>

#include "simulator.h"
#include "simulator_internals.h"
#include "viewer.h"
#include "debug.h"

static int g2_dev;	
static double last_robot_x, last_robot_y, last_robot_heading;	/* last robot situation */
static double w_en, h_en;
static double mulx, muly;	/* inverse transformation of coordinates */

/* when you do -number/number, you get less than -1.0 and asin fails. thus this function */
double acos_safe(double x)
{
	if (x < -1.0) x = -1.0;
	else if (x > 1.0) x = 1.0;
	return acos(x);
}

/* shows the window, draws the borders of environment */
void viewer_open(int win_w, int win_h, double env_w, double env_h)
{
	w_en = env_w; 
	h_en = env_h;
	g2_dev = g2_open_X11X(win_w, win_h, 0, 0, "Cargo", "cargo", NULL, 0, 0);
	g2_set_coordinate_system(g2_dev, 10.0, 10.0, (win_w - 20.0) / env_w, (win_h - 120.0) / env_w);
	muly = env_h / (win_h - 120.0);
	mulx = env_w / (win_w - 20.0);
	g2_pen(g2_dev, COLOR_ENV_BORDER);
	g2_rectangle(g2_dev, 0.0, 0.0, env_w, env_h);
	/* display legend */
	g2_pen(g2_dev, COLOR_ROBOT);
	g2_filled_rectangle(g2_dev, 0.0, env_h + muly * 20.0, 10.0 * mulx, env_h + muly * 10.0);
	g2_string(g2_dev, 20.0 * mulx, env_h + muly * 10.0, "robot");
	g2_pen(g2_dev, COLOR_OBSTACLE);
	g2_filled_rectangle(g2_dev, 100.0 * mulx, env_h + muly * 20.0, 110.0 * mulx, env_h + muly * 10.0);
	g2_string(g2_dev, 120.0 * mulx, env_h + muly * 10.0, "obstacle");
	g2_pen(g2_dev, COLOR_MARK);
	g2_filled_rectangle(g2_dev, 0.0, env_h + muly * 40.0, 10.0 * mulx, env_h + muly * 30.0);
	g2_string(g2_dev, 20.0 * mulx, env_h + muly * 30.0, "floor mark");
	g2_pen(g2_dev, COLOR_ACTIVE);
	g2_filled_rectangle(g2_dev, 100.0 * mulx, env_h + muly * 40.0, 110.0 * mulx, env_h + muly * 30.0);
	g2_string(g2_dev, 120.0 * mulx, env_h + muly * 30.0, "active area");
	
}

/* closes the window */
void viewer_close()
{
	g2_close(g2_dev);
}

/* draws a rectangular obstacle */
static void obstacle(double x, double y, double w, double h)
{
	g2_pen(g2_dev, COLOR_OBSTACLE);
	g2_filled_rectangle(g2_dev, x, y, x + w, y + h);
	g2_pen(g2_dev, 1);
	g2_rectangle(g2_dev, x, y, x + w, y + h);
}

/* draws a rectangular floor mark */
static void mark(double x, double y, double w, double h, int type)
{
	g2_pen(g2_dev, COLOR_MARK);		// here we should use the type as well
	g2_filled_rectangle(g2_dev, x, y, x + w, y + h);
}

/* draws a rectangular active area */
static void active_area(double x, double y, double w, double h)
{
	g2_pen(g2_dev, COLOR_ACTIVE);
	g2_filled_rectangle(g2_dev, x, y, x + w, y + h);
}

/* draws a circular robot */
static void robot(double x, double y, double r, double heading)
{
	g2_pen(g2_dev, COLOR_ROBOT);
	g2_filled_circle(g2_dev, x, y, robot_radius);
	g2_pen(g2_dev, 1);
	g2_circle(g2_dev, x, y, robot_radius);
	g2_line(g2_dev, x, y, x + robot_radius * sin(heading), y + robot_radius * cos(heading));	
}

//converts angle from radians to degrees
static double rad2deg(double rad)
{
	return rad / pi * 180.0;
}

/* based on the simulator data structures, draws the environment contents */
void viewer_refresh()
{
	int i;
	char s[40];
	double alpha1, alpha2;
	
	if ((robot_x == last_robot_x) && (robot_y == last_robot_y) &&
		(robot_heading == last_robot_heading))
		return;

	last_robot_x = robot_x;
	last_robot_y = robot_y;
	last_robot_heading = robot_heading;
	
	/* repaint the environment area */
	g2_pen(g2_dev, COLOR_ENV_BACKGROUND);
	g2_filled_rectangle(g2_dev, 0.0, 0.0, w_en, h_en);

	/* draw all active areas */
	for (i = 0; i < activecnt; i++)
	  active_area(active_x[i], active_y[i], active_w[i], active_h[i]);

	/* draw all floor marks */
	for (i = 0; i < markcnt; i++)
	  mark(mark_x[i], mark_y[i], mark_w[i], mark_h[i], mark_value[i]);

	/* draw all obstacles */
	for (i = 0; i < obstcnt; i++)
	  obstacle(obs_x[i], obs_y[i], obs_w[i], obs_h[i]);

	/* draw robot trajectory */
	g2_pen(g2_dev, COLOR_TRAJECTORY);
	for (i = 0; i < traji - 1; i++)
		if ((traj_moving[i] == MOVING_FWD_STRAIGHT) || 
		    (traj_moving[i] == MOVING_BWD_STRAIGHT))
		    g2_line(g2_dev, traj_x[i], traj_y[i], traj_x[i + 1], traj_y[i + 1]);
		else
		{
		    alpha1 = acos_safe((traj_x[i] - traj_c_x[i]) / traj_c_r[i]);
			alpha2 = acos_safe((traj_x[i + 1] - traj_c_x[i]) / traj_c_r[i]);

			debugmsg2double(DL_DEBUGGING, "alpha1: %lf, alpha2: %lf", alpha1, alpha2);
			debugmsg2double(DL_DEBUGGING, "traj_x: %lf, traj_c_x: %lf", traj_x[i], traj_c_x[i]);
			
			if (traj_y[i] < traj_c_y[i]) alpha1 = 2 * pi - alpha1;
			if (traj_y[i + 1] < traj_c_y[i]) alpha2 = 2 * pi - alpha2; 

			if ((traj_moving[i] == MOVING_FWD_RIGHT) || (traj_moving[i] == MOVING_BWD_LEFT))
				g2_arc(g2_dev, traj_c_x[i], traj_c_y[i], traj_c_r[i], traj_c_r[i], rad2deg(alpha2), rad2deg(alpha1));
			else 
				g2_arc(g2_dev, traj_c_x[i], traj_c_y[i], traj_c_r[i], traj_c_r[i], rad2deg(alpha1), rad2deg(alpha2));
	
		}

	/* draw a robot */
	robot(robot_x, robot_y, robot_radius, robot_heading);

	/* border of environment area */
	g2_pen(g2_dev, COLOR_ENV_BORDER);
	g2_rectangle(g2_dev, 0.0, 0.0, w_en, h_en);

	/* clear status area */
	g2_pen(g2_dev, COLOR_ENV_BACKGROUND);
	g2_filled_rectangle(g2_dev, 200.0 * mulx, h_en + 5.0 * muly, 300.0 * mulx, h_en + 80.0 * muly);
	g2_filled_rectangle(g2_dev, 0.0, h_en + 45.0 * muly, 200.0 * mulx, h_en + 60.0 * muly);
	g2_pen(g2_dev, 1);
	
	/* show simulation time */
	sprintf(s, "sim.time: %d.%d", new_simulation_time / 1000, new_simulation_time % 1000);
	g2_string(g2_dev, 200.0 * mulx, h_en + 30.0 * muly, s);

	/* show movement type */
	g2_string(g2_dev, 200.0 * mulx, h_en + 10.0 * muly, moving_desc[robot_moving]);

	/* show robot coordinates */
	sprintf(s, "x: %lf", robot_x);
	g2_string(g2_dev, 0.0, h_en + 50.0 * muly, s);
	sprintf(s, "y: %lf", robot_y);
	g2_string(g2_dev, 100.0 * mulx, h_en + 50.0 * muly, s);
 
	/* show robot speed */
	sprintf(s, "speed: %lf", robot_speed);
	g2_string(g2_dev, 200.0 * mulx, h_en + 50.0 * muly, s);
	
}
